from __future__ import annotations
from os.path import dirname, join
from typing import Any, Callable, Dict, List, Optional

from shapely.geometry import Polygon
import pandas as pd
import pickle

from .location import Location, Locations
from ..geom.shape import Shape

class Mall(Location):
    """
    Object to encapsulate a Mall in Singapore

    Fields:
        floors (Optional[int]): the number of floors in the mall, exclusive of parking and residential floors.
        stores (Optional[int]): the number of stores in the mall, decided based on the mall directory.
        opening_year (Optional[int]): the year the mall opened.
            This is based on the current name of the mall, so Funan and KINEX count as 2019.
        retail_area (Optional[float]): the total retail space available in the mall.
            This only includes shop space and not walkways and escalators etc.
    """
    floors: Optional[int]
    stores: Optional[int]
    opening_year: Optional[int]
    retail_area: Optional[float]

    def __init__(self, name: str, **kwargs):
        fields = ["lat", "lon", "shape", "floors", "stores", "opening_year", "retail_area", "planning_area"]
        self._try_setter(fields, kwargs)
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class Malls(Locations):
    _FIELD_MAP = {
        "Latitude": "lat",
        "Longitude": "lon",
        "Floors": "floors",
        "Stores": "stores",
        "Opening year": "opening_year",
        "Area": "retail_area",
        "shape": "shape",
    }

    def __init__(self, *malls: Mall, name: str="mall"):
        super().__init__(*malls, name=name)

    @staticmethod
    def get(blanks: bool=False, offline: bool=True) -> Malls:
        raw_df = Malls._get_data_handler(offline)
        data_dict = Malls._get_data_cleaning(blanks)(raw_df)
        malls = Malls._get_data_compiling(data_dict)
        return Malls(*malls)

    @staticmethod
    def _get_data_handler(offline: bool) -> pd.DataFrame:
        if offline:
            return pd.read_csv(join(dirname(__file__), "assets/malls.csv"))
        print("Retrieving 'Malls Raw' from Sheets...")
        raw_df = Locations.get_sheet("Malls Raw")
        print("Retrieved.")
        return raw_df

    @staticmethod
    def _get_data_cleaning(blanks: bool) -> Callable[[pd.DataFrame], Dict[str, Any]]:
        if blanks:
            return lambda df: df.set_index("Name").to_dict("index")
        return lambda df: df[pd.notna(df.Floors)
                             & pd.notna(df.Stores)
                             & pd.notna(df.Area)].set_index("Name").to_dict("index")
    
    @staticmethod
    def _get_data_compiling(data_dict: Dict[str, Any]) -> List[Mall]:
        malls: List[Mall] = []
        with open(join(dirname(__file__), "assets/mall_shapes.pickle"), 'rb') as f:
            mall_shapes_dict: Dict[str, Polygon] = pickle.load(f)
        for name, info in data_dict.items():
            shape = Shape.from_polygon(mall_shapes_dict.get(name))
            info["shape"] = shape
            malls.append(Mall(name, **Malls._field_map(info)))
        return malls

    @staticmethod
    def _field_map(d: Dict[str, Any]) -> Dict[str, Any]:
        for old_field, new_field in Malls._FIELD_MAP.items():
            d[new_field] = d.pop(old_field)
        return d



from __future__ import annotations
from functools import cached_property
from typing import List, Generic, Optional, Tuple, TypeVar

from .bound import Bound
from .quick_sort import median_with_left_right
from ..geom.pointable import Pointable

T = TypeVar("T", bound=Pointable)

class XY:
    X, Y = "x", "y"

class KDNode(Generic[T]):
    """
    Encapsulates a node in a KDTree.

    Fields:
        point (T): the point it is representing in the tree.
        level (str): whether the we compare x or y-values at this point.
            (We alternate between x and y).
        left (Optional[KDNode[T]]): the left child of this node.
        right (Optional[KDNode[T]]): the right child of this node.
    """
    point: T
    level: str
    left:  Optional[KDNode[T]]
    right: Optional[KDNode[T]]

    def __init__(self, point: T, level: str):
        """
        Initialiser for the KDNode[T] object.
        The left and right children are both set to zero
            as this node would be considered as a leaf.

        Args:
            point (T): point to be represented by the node.
            level (str): 'x' or 'y'.
        """
        self.point = point
        self.level = level
        self.left = None
        self.right = None
        
    @cached_property
    def next_level(self) -> str:
        """
        Computes the next level of this node.
        So, 'x' will be mapped to 'y' and vice versa.
        This helps to eliminate the need for if statements when traversing the tree.

        Returns:
            str: 'x' or 'y' value.
        """
        if self.level == XY.X:
            return XY.Y
        else:
            return XY.X
        
    def add(self, point: T) -> None:
        """
        Adds the point to the node.
        If there is a vacancy, the point will join the tree as a new node.
        Otherwise, the point will be passed on to the relevant child's add method.

        Args:
            point (T): the point to be added to the node.
        """
        if getattr(point, self.level) <= getattr(self.point, self.level):
            if self.left == None:
                self.left = KDNode[T](point, self.next_level)
            else:
                self.left.add(point)
        else:
            if self.right == None:
                self.right = KDNode[T](point, self.next_level)
            else:
                self.right.add(point)
                
    def nearest(self, point: T) -> Tuple[Optional[T], float]:
        """
        Gets the nearest point to a particular target point.

        Args:
            point (T): target point.

        Returns:
            Tuple[Optional[T], float]: point-distance tuple.
        """
        next_branch:  Optional[KDNode[T]]
        other_branch: Optional[KDNode[T]]
        temp: Optional[T]

        next_branch, other_branch = self._get_next_other_branch(point)
        temp = self._get_temp_point(next_branch, point)
        return self._backtrack(point, temp, other_branch)

    def _get_next_other_branch(self, point: Pointable) -> Tuple[Optional[KDNode[T]], Optional[KDNode[T]]]:
        """
        There is no guarantee that the deepest node in the tree is the closest point.
        As such, we need to do checks on some sibling nodes we encounter along the way.
        This method returns two nodes, the first one being the more important branch,
            which we will traverse all the way.
        For the other branch, we would only need to check its root node.

        Args:
            point (Pointable): target point.

        Returns:
            Tuple[Optional[KDNode[T]], Optional[KDNode[T]]]: roots of the two child branches.
        """
        if getattr(point, self.level) < getattr(self.point, self.level):
            return (self.left, self.right)
        else:
            return (self.right, self.left)

    def _get_temp_point(self, next_branch: Optional[KDNode[T]], point: T) -> Optional[T]:
        """
        Processes the 'next_branch' to return a node if it exists.

        Args:
            next_branch (Optional[KDNode[T]]): branch to process.
            point (T): target point.

        Returns:
            Optional[T]: a point if it exists.
        """
        if next_branch is not None:
            return next_branch.nearest(point)[0]
        else:
            return None

    def _backtrack(self, point: T, temp: Optional[T], other_branch: Optional[KDNode[T]]) -> Tuple[Optional[T], float]:
        """
        This method computes distances to all the other nodes that we have KIVed.
        The minimum distance is the one that we output.

        Args:
            point (T): target point.
            temp (Optional[T]): current closest point to target.
            other_branch (Optional[KDNode[T]]): other branch to check for closest point.

        Returns:
            Tuple[Optional[Pointable], float]: point-distance tuple.
        """
        best:   Optional[T]
        best_d: float
        s_d:    float

        best, best_d = point.get_closest_point(*[temp, self.point])
        s_d = abs(getattr(point, self.level) - getattr(self.point, self.level))

        if best_d >= s_d and other_branch is not None:
            temp = other_branch.nearest(point)[0]
            best, best_d = point.get_closest_point(*[temp, best])# if temp and best else (None, float("inf"))

        return (best, best_d)

class KDTree(Generic[T]):
    """
    Encapsulates a KDTree object, containing many KDNode[T]s.
    At each depth level, we compare alternating coordinates, starting with x.
    
    Fields:
        root (Optional[KDNode[T]]): the root of the tree.
        weight (int): the number of nodes in the tree.
        bound (Bound): the bounds that contain the entire collection of points.
    """
    root: Optional[KDNode[T]]
    weight: int
    bound: Bound[T]

    def __init__(self):
        """
        Initialiser for an empty KDTree.
        Sets the root to be empty, weight to be 0 and bounds to be infinite.
        """
        self.root = None
        self.weight = 0
        self.bound = Bound(float("inf"), float("-inf"), float("inf"), float("-inf"))

    def __str__(self) -> str:
        return str(self.in_order())

    def __repr__(self) -> str:
        return f"<KDTree: {self.bound}>"

    @property
    def center(self) -> Tuple[float, float]:
        """
        Based on the bounds of the KDTree, return its center point.

        Returns:
            T: center of the KDTree.
        """
        return self.bound.center

    def add_all(self, *points: T) -> None:
        """
        From a collection of points, add all of them to the tree.
        It will check whether the root is None (assign root to this new node then).
        If not None, then the point will be added to the tree's root, and passed along there.
        Weights are also updated.
        
        Args:
            *points (T): the points to be added to the tree.
        """
        sorted_points: List[T] = []
        def append_median(_list: List[T], xy: str) -> None:
            if len(_list) == 0:
                return
            left, mid, right = median_with_left_right(_list, comparator=lambda point: getattr(point, xy))
            if mid is not None:
                sorted_points.append(mid)
            if xy == "x":
                append_median(left, "y")
                append_median(right, "y")
            else:
                append_median(left, "x")
                append_median(right, "x")
        append_median(list(points), "x")
        
        if not self.root:
            self.root = KDNode[T](sorted_points.pop(0), "x")
            self.weight += 1
            
        for point in sorted_points:
            self._remap_min_max(point)
            self.root.add(point)
            self.weight += 1

    def _remap_min_max(self, point: T) -> None:
        """
        Updates the bounds of the KDTree, expanding if a point lies outside of it.

        Args:
            point (T): point to be added.
        """
        self.bound.min_x = min(self.bound.min_x, point.x)
        self.bound.min_y = min(self.bound.min_y, point.y)
        self.bound.max_x = max(self.bound.max_x, point.x)
        self.bound.max_y = max(self.bound.max_y, point.y)
        
    def nearest(self, point: T) -> Tuple[Optional[T], float]:
        """
        Finds the nearest point to the target.
        The code passes the 'point' argument to the root, and lets the nodes handle it.
        If an answer is unavailable, return (None, float('inf'))

        Args:
            point (T): point to be queried.

        Returns:
            Tuple[Optional[T], float]: point-distance tuple.
        """
        if self.root == None:
            return (None, float("inf"))
        return self.root.nearest(point)

    def in_order(self) -> List[T]:
        """
        In order traversal of the tree done by recursion.

        Returns:
            List[T]: a list of points in the tree.
        """
        L: List[T] = []
        def in_order_helper(node: Optional[KDNode[T]]) -> None:
            if not node:
                return None
            in_order_helper(node.left)
            L.append(node.point)
            in_order_helper(node.right)
        in_order_helper(self.root)
        return L

    def pre_order(self) -> List[Pointable]:
        """
        Pre order traversal of the tree done by recursion.

        Returns:
            List[Pointable]: a list of points in the tree.
        """
        L: List[Pointable] = []
        def pre_order_helper(node: Optional[KDNode[T]]) -> None:
            if not node:
                return None
            L.append(node.point)
            pre_order_helper(node.left)
            pre_order_helper(node.right)
        pre_order_helper(self.root)
        return L
           


from __future__ import annotations
from os.path import join, dirname
from typing import Any, Dict, List, Optional, TYPE_CHECKING

from shapely import geometry
import fiona
import geopandas as gpd

from ...color.color import Color
from ...geom.geo_pt import GeoPt
from ...geom.line import Line

if TYPE_CHECKING:
    from . import platform

class MRTLine(Line):
    name: str
    color: Color
    platforms: List[platform.Platform]
    
    def __init__(self, name: str, color: Color, line: Optional[List[GeoPt]]):
        super(Line, self).__init__(line)
        self.name = name
        self.color = color
        self.platforms = []
        
    def __str__(self) -> str:
        return f"<MRTLine: {self.name}>"
    
    def __repr__(self) -> str:
        return f"<MRTLine: {self.name}>"
    
def get_mrt_lines() -> Dict[str, MRTLine]:
    MRT_LINES = {}
    
    gpd.io.file.fiona.drvsupport.supported_drivers['KML'] = 'rw'
    lines = gpd.read_file(join(dirname(__file__), "assets/mrt-line.kml"))
    lines.loc[5:8,"geometry"] = lines.loc[5:8, "geometry"].boundary
    lines.drop(columns=["Description"], inplace=True)
    lines.Name = lines.Name.str.extract("([^L]+)")
    lines_dict = lines.set_index("Name").to_dict("index")
    lines_dict["CE"] = {}
    lines_dict["CE"]["geometry"] = geometry.LineString(list(GeoPt(x[1], x[0]) for x in lines_dict["CC"]["geometry"].coords[346:][::-1]))
    
    color_mapping = {
        "EW": "009645",
        "CG": "009645",
        "NS": "D42E12",
        "NE": "9900AA",
        "CC": "FA9E0D",
        "CE": "FA9E0D",
        "DT": "005EC4",
        "TE": "9D5B25",
        "JS": "0099AA",
        "JE": "0099AA",
        "JW": "0099AA",
        "CR": "97C616",
        "CP": "97C616",
        "SE": "748477",
        "SW": "748477",
        "PE": "748477",
        "PW": "748477",
        "BP": "748477",
    }
    
    for name, data in lines_dict.items():
        MRT_LINES[name] = MRTLine(name,
                                  Color.from_hex(color_mapping[name]),
                                  [GeoPt(x[1], x[0]) for x in data["geometry"].coords])
    return MRT_LINES
    


from __future__ import annotations
from dataclasses import dataclass

from ..error.value_error.invalid_hex_error import InvalidHexError

@dataclass(frozen=True)
class Color:
    r: float
    g: float
    b: float

    def get_diff(self, r: float, g: float, b: float) -> float:
        return abs(r-self.r) + abs(g-self.g) + abs(b-self.b)
    
    def get_diff_from_color(self, color: Color) -> float:
        return self.get_diff(color.r, color.g, color.b)
    
    @staticmethod
    def from_hex(hex: str) -> Color:
        if len(hex) != 6:
            raise InvalidHexError(hex)
        r = int(hex[0:2], 16)
        g = int(hex[2:4], 16)
        b = int(hex[4:6], 16)
        return Color(r, g, b)
    
    def to_hex(self) -> str:
        r_hex = hex(int(self.r))[2:]
        if len(r_hex) == 1:
            r_hex = "0" + r_hex
            
        g_hex = hex(int(self.g))[2:]
        if len(g_hex) == 1:
            g_hex = "0" + g_hex
            
        b_hex = hex(int(self.b))[2:]
        if len(b_hex) == 1:
            b_hex = "0" + b_hex
            
        return r_hex + g_hex + b_hex


from __future__ import annotations
from os.path import join, dirname
from typing import Dict, List, Optional

from PIL import Image
import numpy as np

from ..color.color import Color
from ..error.value_error.out_of_bounds_error import OutOfBoundsError

"""
This script maps a set of lat longs in 
    Singapore to a particular elevation, in metres.

1. Screenshot an image of Singapore's relief map.
2. Process it using PIL and numpy.
3. Map specific colors in the image to specific elevations.
4. Given a set of lat longs, normalise them to get the
    appropriate cell within the picture.
5. Use convolution to get a more accurate reading.
"""

class Elevation():
    """
    A node in a doubly-linked list representing an elevation point.
    It is used to mainly find the elevations just above
        and just below it, so that we can more accurately
        judge the elevations in between.
    Fields:
        c (Color): the color representing the elevation point, noted in RGB.
        e (float): the elevation value, noted in metres.
        higher (Elevation): the Elevation object for the
            immediate next elevation.
        lower (Elevation): the Elevation object for the
            immediate previous elevation.
    """

    c:      Color
    e:      float
    higher: Optional[Elevation]
    lower:  Optional[Elevation]

    def __init__(self, c: Color, e: float):
        """
        Initialiser for the Elevation point object.

        Args:
            c (Color): color of the point, encapsulating RGB values.
            e (float): elevation associated with the particular color.
        """
        self.c = c
        self.e = e
        self.higher = None
        self.lower = None
    
    def set_higher(self, other_E: Optional[Elevation]) -> None:
        """
        Sets the Elevation object directly above it.
        This facilitates searching for next values.

        Args:
            other_E (Optional[Elevation]): the higher Elevation point.
        """
        self.higher = other_E
    
    def set_lower(self, other_E: Optional[Elevation]) -> None:
        """
        Sets the Elevation object directly below it.
        This facilitates searching for previous values.

        Args:
            other_E (Optional[Elevation]): the lower Elevation point.
        """
        self.lower = other_E

    def get_diff(self, r: float, g: float, b: float) -> float:
        """
        A metric of how close the given Color is
            to the Elevation's Color, which lets us work out
            intermediate values, instead of giving arbitrarily
            fixed elevations.
        For example, if the color is between that for 10m and 15m,
            then we use this metric to work out a value like 12m.

        Args:
            r (float): red value
            g (float): green value
            b (float): blue value

        Returns:
            float: the metric to figure out intermediate values for elevation.
        """
        return self.c.get_diff(r, g, b)

class ElevationMap(): 
    """
    Generates the ElevationMap to query elevations.
    """
    MIN_LAT = 1.23776
    MIN_LON = 103.61751
    MAX_LAT = 1.47066
    MAX_LON = 104.04360
    D_LAT   = MAX_LAT - MIN_LAT
    D_LON   = MAX_LON - MIN_LON
    IMG_H   = 1030
    IMG_W   = 1885
    _ELEVATIONS: Optional[List[Elevation]] = None
    _PATH_TO_ELEVATION_MAP = join(dirname(__file__), "assets/singapore-elevation.png")
    _IMG = Image.open(_PATH_TO_ELEVATION_MAP)
    _ARR = np.array(_IMG.convert('RGB'))
    
    @classmethod
    def _set_elevations(cls) -> List[Elevation]:
        """
        Private method for initialising the elevation map.

        Returns:
            List[Elevation]: elevation points encapsulating the color-elevation mapping.
        """
        elevations = []
        c_map = {168: Color(255, 255, 255),#
                 163: Color(219, 203, 201),#
                 156: Color(216, 191, 186),#
                 141: Color(217, 153, 151),#
                 133: Color(224, 147, 145),#
                 126: Color(214, 143, 135),#
                 113: Color(218, 162, 132),#
                 105: Color(185, 134, 118),#
                 91: Color(179, 218, 146),#
                 76: Color(202, 220, 139),#
                 73: Color(180, 219, 130),#
                 69: Color(178, 221, 130),#
                 54: Color(167, 231, 139),#
                 50: Color(164, 223, 150),#
                 46: Color(146, 219, 139),#
                 44: Color(149, 220, 142),#
                 43: Color(147, 225, 134),#
                 42: Color(145, 224, 152),#
                 38: Color(148, 225, 171),#
                 36: Color(152, 220, 163),#
                 32: Color(152, 221, 197),#
                 30: Color(146, 219, 197),#
                 29: Color(150, 220, 168),#
                 25: Color(140, 212, 221),#
                 24: Color(150, 220, 195),#
                 23: Color(167, 221, 230),#
                 18: Color(136, 185, 220),#
                 17: Color(148, 191, 223),#
                 16: Color(138, 182, 223),#
                 15: Color(162, 196, 228),#
                 14: Color(153, 196, 241),#
                 12: Color(135, 168, 218),#
                 10: Color(157, 195, 249),#
                 8: Color(137, 168, 216),#
                 7: Color(138, 156, 219),#
                 3: Color(130, 129, 219),#
                 2: Color(143, 136, 222),#
                 0: Color(134, 122, 239),#
                 0: Color(157, 139, 253)#
        }
        
        # Values obtained by color inspection of the original image
        ElevationMap._map_elevations_to_lower(c_map, elevations)
        ElevationMap._map_elevations_to_higher(elevations)

        return elevations

    @staticmethod
    def _map_elevations_to_lower(c_map: Dict[int, Color], elevations: List[Elevation]) -> None:
        """
        For each elevation point, set the lower neighbour.
        This facilitates the searching of previous neighbours.

        Args:
            c_map (Dict[int, Color]): mapping of elevation to its corresponding
                color in the picture.
            elevations (List[Elevation]): elevation points to apply mapping to.
        """
        # Map each node to its lower neighbour
        curr_e: Optional[Elevation] = None
        for key, value in c_map.items():
            e = Elevation(value, key)
            e.set_higher(curr_e)
            curr_e = e
            elevations.append(e)

    @staticmethod
    def _map_elevations_to_higher(elevations: List[Elevation]) -> None:
        """
        For each elevation point, set the higher neighbour.
        This facilitates the searching of next neighbours.

        Args:
            elevations (List[Elevation]): elevation points to apply mapping to.
        """
        # Map each node to its higher neighbour
        curr_e = None
        for e in elevations[::-1]:
            e.set_lower(curr_e)
            curr_e = e

    @classmethod
    def _get_elevations(cls) -> List[Elevation]:
        """
        Runs _set_elevations() if not yet run.
        Else, returns the evaluated result.

        Returns:
            List[Elevation]: elevation points
        """
        if cls._ELEVATIONS == None:
            cls._ELEVATIONS = cls._set_elevations()
        return cls._ELEVATIONS
            
    @staticmethod
    def _get_elevation_from_color(r: float, g: float, b: float) -> float:
        """
        Private method for getting an elevation from rgb values.

        Args:
            r (float): red value.
            g (float): green value.
            b (float): b value.

        Returns:
            float: gets the elevation value associdated with the RGB value.
        """
        # Finds the closest matching Elevation object to the RGB value
        result_e: Optional[Elevation]
        e:        Elevation
        diff:     float

        min_diff = float("inf")
        result_e = None
        for e in ElevationMap._get_elevations():
            diff = e.get_diff(r, g, b)
            if min_diff > diff:
                min_diff = diff
                result_e = e
        
        return ElevationMap._get_elevation_by_diff(result_e, min_diff, r, g, b)

    @staticmethod
    def _get_elevation_by_diff(result_e: Optional[Elevation],
                               min_diff: float,
                               r: float,
                               g: float,
                               b: float) -> float:
        """
        Private method for getting an elevation from rgb values.

        Args:
            result_e (Optional[Elevation]): closest elevation point to the RGB queried.
            min_diff (float): metric for how close the RGB value is to the elevation point.
            r (float): red value.
            g (float): green value.
            b (float): blue value.

        Returns:
            float: elevation.
        """
        # Outputs an elevation based on how "far" the RGB value
        # is from the two surrounding Elevation points
        lower:    float
        higher:   float
        e:        float
        e_lower:  float
        e_higher: float

        lower  = ElevationMap._get_lower_neighbour(result_e, r, g, b)
        higher = ElevationMap._get_higher_neighbour(result_e, r, g, b)

        if not result_e:
            return float("-inf")

        if lower < higher:
            if not result_e.lower:
                return float("inf")
            e = result_e.e
            e_lower = result_e.lower.e
            return (e*lower + e_lower*min_diff)/(lower+min_diff)
        else:
            if not result_e.higher:
                return float("inf")
            e = result_e.e
            e_higher = result_e.higher.e
            return (e*higher + e_higher*min_diff)/(higher+min_diff)

    @staticmethod
    def _get_lower_neighbour(result_e: Optional[Elevation],
                             r: float,
                             g: float,
                             b: float) -> float:
        """
        Gets the closeness to the lower elevation point
            in order to compute intermediate values.

        Args:
            result_e (Optional[Elevation]): elevation point closest to the RGB value queried.
            r (float): red value.
            g (float): green value.
            b (float): blue value.

        Returns:
            float: closeness metric to the lower elevation point.
        """
        # Gets the lower neighbour where possible
        if result_e == None or result_e.lower == None:
            return float("inf")
        else:
            return result_e.lower.get_diff(r, g, b)
    
    @staticmethod
    def _get_higher_neighbour(result_e: Optional[Elevation],
                              r: float,
                              g: float,
                              b: float) -> float:
        """
        Gets the closeness to the higher elevation point
            in order to compute intermediate values.

        Args:
            result_e (Optional[Elevation]): elevation point closest to the RGB value queried.
            r (float): red value.
            g (float): green value.
            b (float): blue value.

        Returns:
            float: closeness metric to the higher elevation point.
        """
        # Gets the higher neighbour where possible
        if result_e == None or result_e.higher == None:
            return float("inf")
        else:
            return result_e.higher.get_diff(r, g, b)
        
    @staticmethod
    def get_elevation(lat: float, lon: float) -> float:
        """
        Public method to get elevation based on lat and lon.

        Args:
            lat (float): latitude.
            lon (float): longitude.

        Raises:
            OutOfBoundsError: point queried lies outside the bounds of the Singapore map.
            ValueError: both latitude and longitude lie outside the bounds of the
                Singapore map.

        Returns:
            float: elevation at that location.
        """
        # Raises ValueError if query is out of bounds of Singapore
        if not ElevationMap.in_singapore(lat, lon):
            if ElevationMap.in_singapore(lat, 103.85):
                raise OutOfBoundsError(lon, ElevationMap.MIN_LON, ElevationMap.MAX_LON)
            elif ElevationMap.in_singapore(1.35, lon):
                raise OutOfBoundsError(lon, ElevationMap.MIN_LAT, ElevationMap.MAX_LAT)
            raise ValueError("Both coordinates out of bounds!")
        
        # Normalises the lat long to represent cells in the RGB array
        normalised = ElevationMap.normalise_latlong(lat, lon)
        
        # Convolutes the values for a more accurate result
        return ElevationMap._convolute(normalised[0], normalised[1], 2)
    
    @staticmethod
    def in_singapore(lat: float, lon: float) -> bool:
        """
        Checks whether a point is in Singapore, based on the queried
            latitude and longitude.

        Args:
            lat (float): latitude.
            lon (float): longitude.

        Returns:
            bool: whether the point is in Singapore.
        """
        return (lat <= ElevationMap.MAX_LAT and 
                lat >= ElevationMap.MIN_LAT and 
                lon <= ElevationMap.MAX_LON and 
                lon >= ElevationMap.MIN_LON)

    @staticmethod
    def _convolute(row: int, col: int, deg: int) -> float:
        """
        Private method for getting an average
            of a square around the queried location
            in order to get a more accurate reading of elevation.

        Args:
            row (int): in the picture, select this row of pixels.
            col (int): in the picture, select this column of pixels.
            deg (int): the size of the window to be convoluted.

        Returns:
            float: the average elevation amongst the surrounding pixels.
        """
        # Finds the average elevation amongst the cells in the
        # square surrounding the queried cell
        total_e = 0
        for r in range(-deg, deg+1):
            for c in range(-deg, deg+1):
                reflected_row: int = ElevationMap.reflect_row(row+r)
                reflected_col: int = ElevationMap.reflect_col(col+r)
                total_e += ElevationMap._get_elevation_from_color(
                    *ElevationMap._ARR[reflected_row][reflected_col]
                )
        return round(total_e / (deg*2+1)**2, 1)

    @staticmethod
    def reflect_row(r: int) -> int:
        """
        Used to handle the edge cases for convolution,
            done by reflecting the point across the edges of the map.

        Args:
            r (int): row number.

        Returns:
            int: reflected row number if necessary.
        """
        # Maps row values to lie within the bounds
        if r < 0:
            return abs(r)
        elif r >= ElevationMap.IMG_H:
            return (ElevationMap.IMG_H
                    - abs(ElevationMap.IMG_H-r))
        return r

    @staticmethod
    def reflect_col(c: int) -> int:
        """
        Used to handle the edge cases for convolution,
            done by reflecting the point across the edges of the map.

        Args:
            c (int): column number.

        Returns:
            int: reflected column number if necessary.
        """
        # Maps column values to lie within the bounds
        if c < 0:
            return abs(c)
        elif c >= ElevationMap.IMG_W:
            return (ElevationMap.IMG_W
                    - abs(ElevationMap.IMG_W - c))
        return c
    
    @staticmethod
    def normalise_latlong(lat: float, lon: float) -> tuple:
        """
        Converts lat long coordinates into pixels in the picture.

        Args:
            lat (float): latitude.
            lon (float): longitude.

        Returns:
            tuple: coordinate pair for (row, column).
        """
        row = int((lat - ElevationMap.MIN_LAT)
                   * ElevationMap.IMG_H
                   / ElevationMap.D_LAT)
        col = int((lon - ElevationMap.MIN_LON)
                   * ElevationMap.IMG_W
                   / ElevationMap.D_LON)
        return (row, col)

    @staticmethod
    def get_color_from_latlong(lat: float, lon: float) -> Color:
        """
        Gets the color from lat long coordinates.
        This is based on the processed image.

        Args:
            lat (float): latitude.
            lon (float): longitude.

        Returns:
            Color: color value at that pixel in the picture.
        """
        normalised = ElevationMap.normalise_latlong(lat, lon)
        return ElevationMap._ARR[normalised[0]][normalised[1]]



from __future__ import annotations
from os.path import dirname, join
from typing import Callable, Dict, List

import geopandas as gpd
import pandas as pd

from ...locations.location import Location, Locations

class Exit(Location):
    """
    Encapsulates a train station exit.

    Fields:
        exit_code (str): the code given to the station exit.
    """
    exit_code: str

    def __init__(self, name: str, **kwargs):
        fields = ["lat", "lon", "shape", "exit_code"]
        self._try_setter(fields, kwargs)
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class Exits(Locations):
    _FIELD_MAP = {
        "EXIT_CODE_": "exit_code",
    }

    def __init__(self, *exits: Exit, name: str="exit"):
        super().__init__(*exits, name=name)

    @staticmethod
    def get(blanks: bool=False, offline: bool=True) -> Exits:
        raw_df = Exits._get_data_handler(offline)
        data_dict = Exits._get_data_cleaning(blanks)(raw_df)
        exits = Exits._get_data_compiling(data_dict)
        return Exits(*exits)

    @staticmethod
    def _get_data_handler(offline: bool) -> pd.DataFrame:
        raw_df = gpd.read_file(join(dirname(__file__), "assets/Train_Station_Exit_Layer.shp"))
        raw_df["geometry"] = raw_df.geometry.to_crs(epsg=4326)
        return raw_df

    @staticmethod
    def _get_data_cleaning(blanks: bool) -> Callable[[pd.DataFrame], Dict]:
        return lambda df: df.to_dict("index")
    
    @staticmethod
    def _get_data_compiling(data_dict: Dict) -> List[Exit]:
        exits: List[Exit] = []
        for name, info in data_dict.items():
            exits.append(Exit(name, **Exits._field_map(info)))
        return exits

    @staticmethod
    def _field_map(d: Dict) -> Dict:
        for old_field, new_field in Exits._FIELD_MAP.items():
            d[new_field] = d[old_field]
            d.pop(old_field)
        d["lat"] = d["geometry"].y
        d["lon"] = d["geometry"].x
        return d


from __future__ import annotations
from typing import Optional, Tuple

import shapely.geometry

from . import pt
from ..geom.distance import DistanceCalculator
from ..geom.pointable import Pointable
from ..structures.bound import Bound

class GeoPt(shapely.geometry.Point, Pointable):
    """
    Encapsulates a point on the Earth's surface, involving different computations for distances.

    Fields:
        lat (float): latitude of the point.
        lon (float): longitude of the point.
    """
    lat: float
    lon: float

    def __init__(self, lat: float, lon: float):
        """
        Initialiser for the GeoPt method.

        Args:
            lat (float): latitude of the point.
            lon (float): longitude of the point.
        """
        self.lat = lat
        self.lon = lon
        super().__init__(lon, lat)
    
    def __str__(self) -> str:
        return f"GEOPT ({self.lat}, {self.lon})"
        
    def __repr__(self) -> str:
        return f"({self.lat}, {self.lon})"
    
    @property
    def x(self) -> float:
        return self.lon
    
    @property
    def y(self) -> float:
        return self.lat
    
    @staticmethod
    def from_bound(bound: Bound[GeoPt]) -> GeoPt:
        return GeoPt((bound.max_y+bound.min_y)/2, (bound.max_x+bound.min_x)/2)
    
    @staticmethod
    def from_point(point: shapely.geometry.Point) -> GeoPt:
        """
        Factory method that converts a geometry.Point object into a GeoPt object.

        Args:
            point (geometry.Point): point to be converted into GeoPt object.

        Returns:
            GeoPt: the returned GeoPt object.
        """
        return GeoPt(point.y, point.x)
    
    def as_pt(self) -> pt.Pt:
        """
        Converts the GeoPt into a Pt object.

        Returns:
            Pt: the returned Pt.
        """
        return pt.Pt(self.x, self.y)
    
    def coords_as_tuple_latlong(self) -> Tuple[float, float]:
        """
        Converts the point into a lat-long coordinate tuple.

        Returns:
            Tuple[float, float]: lat-long coordinate tuple.
        """
        return (self.lat, self.lon)
    
    def get_distance(self, point: GeoPt) -> float:
        """
        Computes the distance to another GeoPt object.
        Does not allow computing distance to Pt objects as they involve different computations.

        Args:
            point (GeoPt): target point to be computed.

        Raises:
            TypeError: a Pt was sent in as the point argument.

        Returns:
            float: distance, in metres, to the target point.
        """
        if hasattr(point, "lat"):
            return DistanceCalculator.get_distance(self, point)
        raise TypeError("Must be a GeoPt!")
        
    def get_distance_basic(self, point: GeoPt) -> float:
        """
        Computes an approximate distance between the two geo points.
        More accurate when distances are small, will be less accurate as the distance increase.
        This is used mainly for comparing distances, instead of acting as the final result.

        Args:
            point (GeoPt): target point to be computed.

        Returns:
            float: approximate distance, in units.
        """
        return 10000*((self.y-point.y)**2+(self.x-point.x)**2)**0.5
    
    def get_closest_point(self, *points: Optional[GeoPt]) -> Tuple[Optional[GeoPt], float]:
        """
        Based on a collection of points, find the nearest to self.
        
        Args:
            *points (Optional[GeoPt]): we will find the closest of these points to self.

        Returns:
            Tuple[Optional[GeoPt], float]: point-distance tuple.
        """
        nearest_point = None
        nearest_dist = float("inf")
        for point in points:
            if point is not None:
                dist = self.get_distance_basic(point)
                if nearest_dist > dist:
                    nearest_dist = dist
                    nearest_point = point
        if not nearest_point:
            return (None, nearest_dist)
        return (nearest_point, self.get_distance(nearest_point))
    
    def move_to(self, new_x: float, new_y: float) -> GeoPt:
        return GeoPt(new_y, new_x)


import geopandas as gpd
import fiona
from ..geom.line import Line

# class Road(Line):
#     gpd.io.file.fiona.drvsupport.supported_drivers["KML"] = "rw"
#     roads = gpd.read_file()

# class RoadNetwork:
#     _PATH_TO_ROAD_NETWORK = join(dirname(__file__), "assets/singapore-elevation.png")
    
#     @classmethod
#     def _get_roads(cls) -> Dict[str, Road]:
#         pass



from .out_of_bounds_error import OutOfBoundsError

class OutOfSingaporeError(OutOfBoundsError):
    MIN_LAT = 1.23776
    MIN_LON = 103.61751
    MAX_LAT = 1.47066
    MAX_LON = 104.04360
    def __init__(self, lat, lon):
        if lat < OutOfSingaporeError.MIN_LAT or lat > OutOfSingaporeError.MAX_LAT:
            super().__init__(lat, OutOfSingaporeError.MIN_LAT, OutOfSingaporeError.MAX_LAT)
        elif lon < OutOfSingaporeError.MIN_LON or lon > OutOfSingaporeError.MAX_LON:
            super().__init__(lon, OutOfSingaporeError.MIN_LON, OutOfSingaporeError.MAX_LON)
        else:
            pass


from __future__ import annotations
from os.path import dirname, join
from typing import Any, Callable, Dict, List, Optional

from shapely.geometry import Polygon
import pandas as pd
import pickle

from .location import Location, Locations
from ..geom.shape import Shape

class School(Location):
    """
    Object to encapsulate a school in Singapore.

    Fields:
        code (Optional[int]): the code number assigned to the school, usually a 4 digit number.
        funding (str): whether the school is government funded etc.
        level (str): whether the school is Primary, Secondary or Tertiary level.
        opening_year (Optional[int]): the year the school was opened.
            If the school has been renamed/merged, choose that year instead.
        gender (str): whether the school is mixed, girls or boys.
    """
    code:         Optional[int]
    funding:      str
    level:        str
    opening_year: Optional[int]
    gender:       str

    def __init__(self, name: str, **kwargs):
        fields = ["lat", "lon", "shape", "code", "funding", "level", "opening_year", "gender"]
        self._try_setter(fields, kwargs)
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class PrimarySchool(School):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)

class SecondarySchool(School):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        fields = ["exp", "na", "nt", "ip"]
        self._try_setter(fields, kwargs)

class TertiarySchool(School):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)

class Schools(Locations):
    _FIELD_MAP = {
        "Latitude": "lat",
        "Longitude": "lon",
        "School Code": "code",
        "Funding": "funding",
        "Level": "level",
        "Opening Year": "opening_year",
        "Type": "gender",
        "shape": "shape",
    }

    def __init__(self, *schools: School, name="school"):
        super().__init__(*schools, name=name)

    @staticmethod
    def get(blanks=False, offline=True) -> Schools:
        raw_df = Schools._get_data_handler(offline)
        data_dict = Schools._get_data_cleaning(blanks)(raw_df)
        schools = Schools._get_data_compiling(data_dict)
        return Schools(*schools)

    @staticmethod
    def _get_data_handler(offline: bool) -> pd.DataFrame:
        if offline:
            return pd.read_csv(join(dirname(__file__), "assets/schools.csv"))
        print("Retrieving 'Schools' from Sheets...")
        raw_df = Locations.get_sheet("Schools")
        print("Retrieved.")
        return raw_df

    @staticmethod
    def _get_data_cleaning(blanks: bool) -> Callable[[pd.DataFrame], List[Dict[str, Any]]]:
        if blanks:
            return lambda df: df.to_dict("records")
        def clean_df(df):
            filtered_df: pd.DataFrame = df[df.Latitude != 0]
            return filtered_df.to_dict("records")
        return clean_df

    @staticmethod
    def _get_data_compiling(data_dict: List[Dict[str, Any]]) -> List[School]:
        schools: List[School] = []
        with open(join(dirname(__file__), "assets/school_shapes.pickle"), 'rb') as f:
            school_shapes_dict: Dict[str, Polygon] = pickle.load(f)
        for school_info in data_dict:
            name = school_info["Name"]
            shape = Shape.from_polygon(school_shapes_dict.get(name))
            school_info["shape"] = shape
            info = Schools._field_map(school_info)
            level = info["level"]
            if level == "Primary":
                schools.append(PrimarySchool(name, **info))
            elif level == "Secondary":
                schools.append(SecondarySchool(name, **info))
            elif level == "Tertiary":
                schools.append(TertiarySchool(name, **info))
        return schools

    @staticmethod
    def _field_map(d: Dict[str, Any]) -> Dict[str, Any]:
        for old_field, new_field in Schools._FIELD_MAP.items():
            d[new_field] = d.pop(old_field)
        return d


from __future__ import annotations
from abc import abstractmethod
from typing import Any, Protocol

class Comparable(Protocol):
    @abstractmethod
    def __lt__(self: Any, other: Any) -> bool:
        pass


from __future__ import annotations
from dataclasses import dataclass, field
from functools import cached_property
from typing import Optional

import numpy as np

# from .platform_code import PlatformCode
from . import connection, platform_code
from ...locations.location import Location

class Platform(Location):
    platform_code: platform_code.PlatformCode
    opening_year: Optional[int]
    closing_year: Optional[int]
    lower_connection: Optional[connection.Connection] = field(default=None, init=False)
    upper_connection: Optional[connection.Connection] = field(default=None, init=False)
    
    def __init__(self, name: str, **kwargs):
        fields = ["lat", "lon", "shape", "platform_code", "opening_year", "closing_year"]
        self._try_setter(fields, kwargs)
        self.lower_connection = None
        self.upper_connection = None
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)
        
    def __eq__(self, other: Platform):
        return self.platform_code == other.platform_code
    
    def __lt__(self, other: Platform):
        return self.platform_code < other.platform_code
    
    def is_in_service(self, year: int) -> bool:
        return ((self.opening_year is None or np.isnan(self.opening_year) or self.opening_year <= year)
                and (self.closing_year is None or np.isnan(self.closing_year) or year <= self.closing_year))
    
    @property    
    def upper_neighbour(self) -> Optional[Platform]:
        if self.upper_connection is not None:
            return self.upper_connection.upper
        return None

    @property
    def lower_neighbour(self) -> Optional[Platform]:
        if self.lower_connection is not None:
            return self.lower_connection.lower
        return None


from typing import Dict, List

from .mrt.connection import Connection
from .mrt.platform import Platform
from .mrt.mrt_line import MRTLine, get_mrt_lines
from .mrt.platform_code import PlatformCode
from .mrt.station import Station, Stations
from ..geom.geo_pt import GeoPt
from ..geom.line import Line

class Network:
    year: int
    station_dict: Dict[PlatformCode, Station]
    platform_dict: Dict[PlatformCode, Platform]
    stations: Stations
    lines: Dict[str, MRTLine]
    
    def __init__(self, year: int=2022):
        self.year = year
        self.stations = Stations.get()
        self.lines = get_mrt_lines()
        self._map_station_dict()
        self._connect_lines()
        
    def _map_station_dict(self) -> None:
        self.station_dict = {}
        self.platform_dict = {}
        for _, station in self.stations.locations.items():
            if isinstance(station, Station):
                for platform in station.platforms:
                    if platform.is_in_service(self.year):
                        self.station_dict[platform.platform_code] = station
                        self.platform_dict[platform.platform_code] = platform
                        prefix = platform.platform_code.prefix
                        if prefix in self.lines:
                            self.lines[prefix].platforms.append(platform)
                        
    def _get_segments(self, line_name):
        line = self.lines[line_name]
        platforms = sorted(line.platforms)

        # Get the closest projections onto the line
        projections: List[GeoPt] = [GeoPt(1,2)]
        for platform in platforms:
            nearest_point = line.get_nearest(platform)[0]
            if nearest_point:
                projections.append(nearest_point)

        segments = []
        curr_start = 0
        for i in range(len(projections)-1):
            lower_proj = projections[i]
            upper_proj = projections[i+1]
            curr_segment = [lower_proj]

            for j in range(curr_start, len(line.coords)-1):
                p1 = line.coords[j]
                p2 = line.coords[j+1]
                min_lat = min(p1[1], p2[1])
                min_lon = min(p1[0], p2[0])
                max_lat = max(p1[1], p2[1])
                max_lon = max(p1[0], p2[0])
                curr_segment.append(GeoPt(p1[1], p1[0]))
                if min_lat <= upper_proj.lat <= max_lat and min_lon <= upper_proj.lon <= max_lon:
                    curr_start = j+1
                    break

            curr_segment.append(upper_proj)

            if i != 0:
                segments.append(curr_segment)
        
        return segments
    
    def _connect_lines(self) -> None:
        for line_name, line in self.lines.items():
            segments = self._get_segments(line_name)
            platforms = sorted(line.platforms)
            for i in range(len(platforms) - 1):
                lower = platforms[i]
                upper = platforms[i+1]
                segment = segments[i]
                connection = Connection(lower, upper, line, segment)
                lower.upper_connection = connection
                upper.lower_connection = connection
    
    @property
    def platforms(self) -> List[Platform]:
        return list(self.platform_dict.values())
            
    def neighbours(self, station: Station) -> List[Station]:
        _list: List[Station] = []
        for platform in station.platforms:
            if platform.upper_neighbour:
                upper_code = platform.upper_neighbour.platform_code
                _list.append(self.station_dict[upper_code])
            if platform.lower_neighbour:
                lower_code = platform.lower_neighbour.platform_code
                _list.append(self.station_dict[lower_code])
        return _list


from __future__ import annotations
from typing import Generic, get_args, Tuple, TypeVar

import shapely.geometry

from ..geom.pointable import Pointable

T = TypeVar("T", bound=Pointable)

class Bound(Generic[T]):
    """
    Encapsulates a rectangular box, possessing coordinates for each of its sides.

    Fields:
        min_x (float): left bound of the rectangular box.
        max_x (float): right bound of the rectangular box.
        min_y (float): top bound of the rectangular box.
        max_y (float): bottom bound of the rectangular box.
    """
    min_x: float
    max_x: float
    min_y: float
    max_y: float

    def __init__(self, min_x: float, max_x: float, min_y: float, max_y: float):
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def __str__(self) -> str:
        return f"x({self.min_x}, {self.max_x}), y({self.min_y}, {self.max_y})"
    
    def merge_with(self, bound: Bound[T]) -> None:
        """
        Merges two bounds together, such that the new bound contains both bounds.

        Args:
            bound (Bound): the bound to be added.
        """
        self.min_x = min(self.min_x, bound.min_x)
        self.max_x = max(self.max_x, bound.max_x)
        self.min_y = min(self.min_y, bound.min_y)
        self.max_y = max(self.max_y, bound.max_y)
        
    @property
    def center(self) -> Tuple[float, float]:
        return ((self.max_x+self.min_x)/2, (self.max_y+self.min_y)/2)
    
    def contains(self, point: T) -> bool:
        """
        Checks whether a point lies within this bound.

        Args:
            point (T): point to be queried.

        Returns:
            bool: whether the point lies within the bound.
        """
        return (point.x >= self.min_x
                and point.x <= self.max_x
                and point.y >= self.min_y
                and point.y <= self.max_y)

    @staticmethod
    def get_bound_from_shape(shape: shapely.geometry.polygon.Polygon) -> Bound:
        """
        Converts a shape into a Bound.

        Args:
            shape (geometry.polygon.Polygon): shape to be converted into bound.

        Returns:
            Bound: the bounds of the shape.
        """
        return Bound[T](shape.bounds[0],
                        shape.bounds[2],
                        shape.bounds[1],
                        shape.bounds[3])

    


from typing import Any, Generic, Optional, TypeVar

from .avltree import AVLTree
from .comparable import Comparable

T = TypeVar("T", bound=Comparable)

class PriorityQueue(Generic[T]):
    """
    Encapsulates a PriorityQueue to enq and deq items in O(logN) time.

    Fields:
        tree (AVLTree[T]): the tree to keep track of the items.
    """
    tree: AVLTree[T]
    
    def __init__(self):
        self.tree = AVLTree[T]()
        
    def is_empty(self) -> bool:
        return self.tree.root == None
    
    def enq(self, val: T) -> None:
        self.tree.insert(val)
        
    def deq(self) -> Optional[T]:
        min_val: Optional[T] = self.tree.get_min_value()
        self.tree.delete(min_val)
        return min_val


from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, Tuple

class Pointable(ABC):
    x: float
    y: float
    
    @abstractmethod
    def __init__(self, x: float, y: float):
        pass
    
    @abstractmethod
    def get_closest_point(self, *points: Pointable) -> Tuple[Optional[Pointable], float]:
        pass
    
    @abstractmethod
    def move_to(self, new_x: float, new_y: float) -> Pointable:
        pass
    
    @abstractmethod
    def get_distance(self, Pointable) -> float:
        pass
        


from __future__ import annotations
from os.path import dirname, join
from typing import Any, Callable, Dict, List, Optional

import pandas as pd

from .location import Location, Locations
from ..geom.elevation import ElevationMap

class Halal(Location):
  """
  Object to encapsulate a Halal store in Singapore.

  Fields:
    store (str): the name of the franchise or store.
    unit (Optional[str]): the unit number of the halal store.
    food_types (str): the type of food they serve.
    cuisines (str): the cuisines of this store.
    foods (str): the food sold in this store.
    location (Optional[Location]): the location of this store.
  """
  store: str
  unit: Optional[str]
  food_types: str
  cuisines: str
  foods: str
  location: Optional[Location]

  def __init__(self, name: str, **kwargs):
    fields = ["lat", "lon", "shape", "store", "unit", "food_types", "cuisines", "foods", "location"]
    self._try_setter(fields, kwargs)
    super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class Halals(Locations):
  _FIELD_MAP = {
    "Latitude": "lat",
    "Longitude": "lon",
    "Unit": "unit",
    "Location": "location"
  }

  def __init__(self, *halals: Halal, name: str="halal"):
    super().__init__(*halals, name=name)

  @staticmethod
  def get(blanks: bool=False, offline: bool=True) -> Halals:
    raw_df = Halals._get_data_handler(offline)
    data_dict = Halals._get_data_cleaning(blanks)(raw_df)
    halals = Halals._get_data_compiling(data_dict)
    return Halals(*halals)

  @staticmethod
  def _get_data_handler(offline: bool) -> pd.DataFrame:
    if offline:
      return pd.read_csv(join(dirname(__file__), "assets/halals.csv"))
    print("Retrieving 'Halals' from Sheets...")
    raw_df = Locations.get_sheet("Halal")
    print("Retrieved.")
    return raw_df

  @staticmethod
  def _get_data_cleaning(blanks: bool=True) -> Callable[[pd.DataFrame], Dict[str, Any]]:
    def combined_name(df: pd.DataFrame) -> pd.DataFrame:
      df["store"] = df.Name
      df["Name"] = df.store + " (" + df.Location + ")"
      return df
    def drop(df: pd.DataFrame) -> pd.DataFrame:
      df = df.drop_duplicates(subset=["Name"], keep="first")
      return df
    if blanks:
      return lambda df: drop(combined_name(df))\
        .set_index("Name").to_dict("index")
    return lambda df: drop(combined_name(df))\
      .set_index("Name").to_dict("index")
      
  @staticmethod
  def _get_data_compiling(data_dict: Dict[str, Any]) -> List[Halal]:
    halals: List[Halal] = []
    for name, info in data_dict.items():
      search: List[str] = info["Search"].split(",")
      food_types: List[str] = []
      cuisines: List[str] = []
      foods: List[str] = []
      dummy = []
      
      list_to_add = food_types
      for i in range(1, len(search)):
        elem = search[i].strip()
        if elem == "Cuisine":
          list_to_add = cuisines
        elif elem == "Food":
          list_to_add = dummy
        elif elem == "Recommend food":
          list_to_add = foods
        else:
          list_to_add.append(elem)
        
      info["food_types"] = ", ".join(food_types)
      info["cuisines"] = ", ".join(cuisines)
      info["foods"] = ", ".join(foods)
      
      if ElevationMap.in_singapore(info["Latitude"], info["Longitude"]):
        halals.append(Halal(name, **Halals._field_map(info)))
    return halals

  @staticmethod
  def _field_map(d: Dict[str, Any]) -> Dict[str, Any]:
    for old_field, new_field in Halals._FIELD_MAP.items():
      d[new_field] = d.pop(old_field)
    return d



from __future__ import annotations
from os.path import join, dirname
from typing import Any, Callable, Dict, List, Optional, Tuple

import geopandas as gpd
import pandas as pd

from .connection import Connection
from .exit import Exits
from .platform import Platform
from .platform_code import PlatformCode
from ...locations.location import Location, Locations
from ...structures.kdtree import KDTree

class Station(Location):
    """
    Station in the MRT network.
    """
    platforms: List[Platform]
    abbr: Optional[str]
    address: Optional[str]
    postcode: Optional[str]
    chinese: Optional[str]
    
    def __init__(self, name: str, platforms: List[Platform], **kwargs):
        fields = ["lat", "lon", "shape", "abbr", "address", "postcode", "chinese"]
        self._try_setter(fields, kwargs)
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)
        self.platforms = platforms

class Stations(Locations):
    stations_by_year: KDTree[Location]

    _FIELD_MAP = {
        "Abbreviation": "abbr",
        "Opening Year": "opening_year",
        "Address": "address",
        "Postcode": "postcode",
        "Lat": "lat",
        "Long": "lon",
        "Closing Year": "closing_year",
        "Chinese": "chinese",
    }

    def __init__(self,
                 *stations: Station,
                 name: str="station",
                 platforms: Dict[str, Station]={}):
        super().__init__(*stations, name=name)
        self.platforms = platforms

    @staticmethod
    def get(blanks: bool=True, offline: bool=True) -> Stations:
        raw_df = Stations._get_data_handler(offline)
        data_dict = Stations._get_data_cleaning(blanks)(raw_df)
        stations = Stations._get_data_compiling(data_dict)
        return Stations(*stations)

    @staticmethod
    def _get_data_handler(offline: bool) -> pd.DataFrame:
        raw_df = pd.read_csv(join(dirname(__file__), "assets/mrt.csv"))
        return raw_df

    @staticmethod
    def _get_data_cleaning(blanks: bool) -> Callable[[pd.DataFrame], List[Dict[str, Any]]]:
        if blanks:
            return lambda df: df.to_dict("records")
        def clean_df(df):
            filtered_df: pd.DataFrame = df[pd.notnull(df.Abbreviation)
                                           & pd.notnull(df["Opening Year"])
                                           & pd.notnull(df.Address)
                                           & pd.notnull(df.Postcode)]
            return filtered_df.to_dict("records")
        return clean_df
    
    @staticmethod
    def _get_data_compiling(data_dict: List[Dict[str, Any]]) -> List[Station]:
        station_platforms: Dict[str, List[Platform]] = {}
        station_infos: Dict[str, Dict[str, Any]] = {}
        stations: List[Station] = []
        
        for platform_info in data_dict:
            name = platform_info["Name"]
            platform_code = PlatformCode(platform_info["Label"])
            platform_info = Stations._field_map(platform_info)
            platform_info["platform_code"] = platform_code
            # lat = platform_info.get("lat")
            # lon = platform_info.get("lon")
            # opening_year = platform_info.get("opening_year")
            # closing_year = platform_info.get("closing_year")
            # platform = Platform(platform_code, opening_year, closing_year)
            platform = Platform(platform_code.code, **platform_info)
            if name not in station_platforms:
                station_platforms[name] = [platform]
                station_infos[name] = platform_info
            else:
                station_platforms[name].append(platform)
                
        for name, platforms in station_platforms.items():
            stations.append(Station(name, platforms, **station_infos[name]))
        return stations

    @staticmethod
    def _field_map(d: Dict) -> Dict:
        for old_field, new_field in Stations._FIELD_MAP.items():
            d[new_field] = d.pop(old_field)
        return d



from __future__ import annotations
from typing import Dict, List, Optional, Tuple

from shapely import geometry
from shapely.ops import nearest_points

from .geo_pt import GeoPt
from .pt import Pt
from ..structures.bound import Bound
from ..structures.kdtree import KDTree

class Shape(geometry.polygon.Polygon):
    """
    This class encapsulates a Shape object, with points represented
        as a KDTree to facilitate the finding of nearest points.

    Fields:
        points: a KDTree full of points representing the Shape.
    """
    points: KDTree[GeoPt]

    def __init__(self, points: List[GeoPt]):
        """
        Initialiser for the Shape object.

        Args:
            points (List[GeoPt]): ordered list of points to be included in the shape.
        """
        self.points = KDTree[GeoPt]()
        super().__init__(points)
        if isinstance(points, list):
            self.points.add_all(*points)

    @staticmethod
    def from_polygon(polygon: Optional[geometry.polygon.Polygon]) -> Optional[Shape]:
        if not polygon:
            return None
        if polygon.exterior:
            return Shape([GeoPt(coords[1], coords[0]) for coords in polygon.exterior.coords])
        return None
        
    def get_nearest(self, point: GeoPt, simple: bool=False) -> Tuple[Optional[GeoPt], float]:
        """
        Gets the nearest point of the Shape to the target.
        Returns 0 if the point is contained within the Shape.

        Args:
            point (GeoPt): target point.

        Returns:
            Tuple[Optional[GeoPt], float]: point-distance pair.
        """
        if simple:
            return (self.center, self.center.get_distance(point))
        if not self.points:
            nearest_point: GeoPt = Pt.from_point(nearest_points(self, point)[0]).as_geo_pt()
            return (nearest_point, point.get_distance(nearest_point))
        if self.contains(point):
            xy = self.points.center
            return (GeoPt(lat=xy[1], lon=xy[0]), 0)
        nearest = self.points.nearest(point)
        return (nearest[0], nearest[1]) if nearest[0] else (None, float("inf"))

    def get_bounds(self) -> Bound:
        return self.points.bound

    @property
    def center(self) -> GeoPt:
        """
        Gets the center of the Shape.
        If it is a geometry.polygon.Polygon object, then find its centroid and return that point.
        Otherwise, just get the center.

        Returns:
            GeoPt: center of the shape.
        """
        if self.points:
            xy = self.points.center
            return GeoPt(xy[1], xy[0])
        return Pt(self.centroid.coords.x, self.centroid.coords.y).as_geo_pt()


from typing import Any, Callable, List, Optional, Tuple, TypeVar

from .comparable import Comparable

T = TypeVar("T")

def swap(arr: List, i: int, j: int) -> None:
    """
    Swaps two entries in a list.

    Args:
        arr (List): array to be swapped.
        i (int): first index to be swapped.
        j (int): second index to be swapped.
    """
    arr[i], arr[j] = arr[j], arr[i]

def partition(arr: List[T], low: int, high: int, comparator: Callable[[T], Any]=lambda x: x) -> int:
    """
    Partition method to assist with Quick Select and Quick Sort algorithms.
    Partition will segment the array based on its last element (pivot).
    Values smaller than the pivot are to its left and
        values greater than the pivot are to its right.
    Comparator can be called for more custom comparisons.

    Args:
        arr (List[T]): array to be partitioned.
        low (int): start index of the subarray to be partitioned.
        high (int): end index of the subarray to be partitioned.
        comparator (T, optional): basis of comparison. Defaults to lambda x: x.

    Returns:
        T: the index of the pivot after everything is shifted.
    """
    i = low - 1
    pivot = arr[high]
 
    for j in range(low, high):
        if not comparator(arr[j]) > comparator(pivot):
            i += 1
            swap(arr, i, j)
 
    swap(arr, i+1, high)
    return i + 1

def quick_sort(arr: List[T],
               comparator: Callable[[T], Any]=lambda x: x
               ) -> List[T]:
    """
    Implementation of the Quick Sort algorithm, accomodating a custom comparator.

    Args:
        arr (List[T]): array to be sorted.
        comparator (T, optional): basis of comparison. Defaults to lambda x: x.

    Returns:
        List[T]: sorted list.
    """
    def quick_sort_helper(low: int, high: int) -> None:
        """
        Helper function for the Quick Sort algorithm.

        Args:
            low (int): start index of the subarray to be sorted.
            high (int): end index of the subarray to be sorted.
        """
        if low >= high:
            return
        p = partition(arr, low, high, comparator)
        quick_sort_helper(low, p-1)
        quick_sort_helper(p+1, high)
    quick_sort_helper(0, len(arr)-1)
    return arr

def quick_select(arr: List[T],
                 n: int,
                 comparator: Callable[[T], Any]=lambda x: x
                 ) -> Optional[T]:
    """
    Implementation of the Quick Select algorithm,
        used to find the nth smallest element in the array.

    Args:
        arr (List[T]): array to be queried.
        n (int): finds the nth smallest element in the array.
        comparator (T, optional): basis of comparison. Defaults to lambda x: x.

    Returns:
        Optional[T]: returns None if array is empty. Otherwise, returns nth smallest element.
    """
    n -= 1
    def quick_select_helper(low: int, high: int) -> Optional[T]:
        """
        Helper function for the Quick Select algorithm.

        Args:
            low (int): start index for the subarray to be queried.
            high (int): end index for the subarray to be queried.

        Returns:
            Optional[T]: recursively calls function on a smaller subarray
                if required element is still not found.
        """
        if low == high == n:
            return arr[n]
        if low > high:
            return None
        p = partition(arr, low, high, comparator)
        if p == n:
            return arr[p]
        elif p > n:
            return quick_select_helper(low, p-1)
        else:
            return quick_select_helper(p+1, high)
    return quick_select_helper(0, len(arr)-1)

def median(arr: List[T],
           comparator: Callable[[T], Any]=lambda x: x
           ) -> Optional[T]:
    """
    Finds the median value of an array.

    Args:
        arr (List[T]): array to be queried.
        comparator (T, optional): basis of comparison. Defaults to lambda x: x.

    Returns:
        Optional[T]: median value, if available.
    """
    return quick_select(arr, int(len(arr)/2)+1, comparator)

def median_with_left_right(arr: List[T],
                           comparator: Callable[[T], T]=lambda x: x
                           ) -> Tuple[List[T], Optional[T], List[T]]:
    """
    Finds the median value, as well as the left and right
        subarrays created after runnning Quick Select.

    Args:
        arr (List[T]): array to be queried.
        comparator (T, optional): basis of comparison. Defaults to lambda x: x.

    Returns:
        Tuple[List[T], Optional[T], List[T]]: median, with left and right subarrays.
    """
    if len(arr) == 0:
        return ([], None, [])
    med = median(arr, comparator)
    mid = int(len(arr)/2)
    return (arr[:mid], med, arr[mid+1:])


from __future__ import annotations
from typing import Any, Callable, Dict, List, Optional, Tuple
from os.path import dirname, join

import fiona
import geopandas as gpd
import pandas as pd
from tqdm import tqdm

from .location import Location, Locations
from ..geom.geo_pt import GeoPt
from ..structures.bounds_tree import BoundsTree

class LandUse(Location):
    """
    Object encapsulating the land use areas in Singapore.

    Fields:
        code (str): the code assigned to the land lot.
    """
    code: str

    def __init__(self, name: str, **kwargs):
        fields = ["lat", "lon", "shape", "code"]
        self._try_setter(fields, kwargs)
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class LandUses(Locations):
    # locations_interval_tree: IntervalTree[LandUse]
    _FIELD_MAP = {
        "geometry": "shape",
    }
    # https://data.gov.sg/dataset/sla-cadastral-land-parcel
    _PATH_TO_LAND_USES = join(dirname(__file__), "assets/sla-cadastral-land-parcel-kml.kml")

    def __init__(self, *areas: LandUse, name="lot"):
        # self.locations_interval_tree = IntervalTree[LandUse]()
        super().__init__(*areas, name=name)
        # self.locations_interval_tree.add_all([area.shape for area in areas if area.shape], [area for area in areas if area])

    @staticmethod
    def get(blanks=False, offline=True) -> LandUses:
        print("Handling...", end=" ")
        raw_df = LandUses._get_data_handler(offline)
        print("Done")
        print("Cleaning...", end=" ")
        data_dict = LandUses._get_data_cleaning(blanks)(raw_df)
        print("Done")
        print("Compiling...", end=" ")
        areas = LandUses._get_data_compiling(data_dict)
        print("Done")
        return LandUses(*areas)

    @staticmethod
    def _get_data_handler(offline: bool) -> pd.DataFrame:
        gpd.io.file.fiona.drvsupport.supported_drivers["KML"] = "rw" # type: ignore [no-redef]
        return gpd.read_file(LandUses._PATH_TO_LAND_USES, driver='KML')

    @staticmethod
    def _get_data_cleaning(blanks: bool) -> Callable[[pd.DataFrame], List[Dict[str, Any]]]:
        def clean_data(df: pd.DataFrame) -> List[Dict[str, Any]]:
            extract_from_description = lambda pattern: df.Description.str.extract(pattern)
            df["Name"] = extract_from_description("<td>(.*?)</td>")
            df = df.drop("Description",axis=1)
            df = df[["Name", "geometry"]]
            return df.to_dict("records")
        return clean_data

    @staticmethod
    def _get_data_compiling(data_dict: List[Dict[str, Any]]) -> List[LandUse]:
        areas: List[LandUse] = []
        for info in tqdm(data_dict):
            name = info.pop("Name")
            areas.append(LandUse(name, **LandUses._field_map(info)))
        return areas

    @staticmethod
    def _field_map(d: Dict[str, Any]) -> Dict[str, Any]:
        for old_field, new_field in LandUses._FIELD_MAP.items():
            if new_field not in d:
                d[new_field] = d.pop(old_field)
        return d

    def get_nearest_to(self, point: GeoPt) -> Tuple[Optional[Location], float]:
        for _, location in self.locations.items():
            if location.shape and location.shape.contains(point):
                return (location, 0)
        return (None, float("inf"))
        


from __future__ import annotations
from typing import Any, Callable, Dict, List, Optional, Tuple
from os.path import dirname, join

import fiona
import geopandas as gpd
import pandas as pd

from .location import Location, Locations
from ..geom.geo_pt import GeoPt
from ..structures.bounds_tree import BoundsTree

class PlanningArea(Location):
    """
    Object encapsulating the planning areas in Singapore.

    Fields:
        code (str): the code assigned to the planning subzone.
        planning_area (str): the medium-level division of Singapore (Woodlands, Tampines, Bedok etc.)
        region (str): the broadest-level division of Singapore (East, West, North etc.)
    """
    code: str
    planning_area: str
    region: str

    def __init__(self, name: str, **kwargs):
        fields = ["lat", "lon", "shape", "code", "planning_area", "region"]
        self._try_setter(fields, kwargs)
        super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class PlanningAreas(Locations):
    interval_tree: BoundsTree[PlanningArea]
    _FIELD_MAP = {
        "geometry": "shape",
        "SubzoneCode": "code",
        "Planning": "planning_area",
        "Region": "region",
    }
    # https://data.gov.sg/dataset/master-plan-2019-subzone-boundary-no-sea
    _PATH_TO_PLANNING_AREAS = join(dirname(__file__), "assets/master-plan-2019-subzone-boundary-no-sea-kml.kml")

    def __init__(self, *areas: PlanningArea, name="planning_area"):
        self.interval_tree = BoundsTree[PlanningArea]()
        super().__init__(*areas, name=name)
        self.interval_tree.add_all([area.shape for area in areas if area.shape], [area for area in areas if area])

    @staticmethod
    def get(blanks=False, offline=True) -> PlanningAreas:
        raw_df = PlanningAreas._get_data_handler(offline)
        data_dict = PlanningAreas._get_data_cleaning(blanks)(raw_df)
        areas = PlanningAreas._get_data_compiling(data_dict)
        return PlanningAreas(*areas)

    @staticmethod
    def _get_data_handler(offline: bool) -> pd.DataFrame:
        gpd.io.file.fiona.drvsupport.supported_drivers["KML"] = "rw" # type: ignore [no-redef]
        return gpd.read_file(PlanningAreas._PATH_TO_PLANNING_AREAS, driver='KML')

    @staticmethod
    def _get_data_cleaning(blanks: bool) -> Callable[[pd.DataFrame], List[Dict[str, Any]]]:
        def clean_data(df: pd.DataFrame) -> List[Dict[str, Any]]:
            extract_from_description = lambda pattern: df.Description.str.extract(pattern)
            # df["SubzoneCode"] = extract_from_description("Subzone Code.*?<td>(.*?)</td>")
            # df["Planning"]    = extract_from_description("Planning Area Name.*?<td>(.*?)</td>")
            # df["Region"]      = extract_from_description("Region Name.*?<td>(.*?)</td>")
            df["Name"]        = extract_from_description("SUBZONE_N</th> <td>(.*?)</td>")
            df["SubzoneCode"] = extract_from_description("SUBZOME_C</th> <td>(.*?)</td>")
            df["Planning"]    = extract_from_description("PLN_AREA_N</th> <td>(.*?)</td>")
            df["Region"]      = extract_from_description("REGION_N</th> <td>(.*?)</td>")
            df = df.drop("Description",axis=1)
            df = df[["Name", "SubzoneCode", "Planning", "Region", "geometry"]]
            return df.to_dict("records")
        return clean_data

    @staticmethod
    def _get_data_compiling(data_dict: List[Dict[str, Any]]) -> List[PlanningArea]:
        areas: List[PlanningArea] = []
        for info in data_dict:
            name = info.pop("Name")
            areas.append(PlanningArea(name, **PlanningAreas._field_map(info)))
        return areas

    @staticmethod
    def _field_map(d: Dict[str, Any]) -> Dict[str, Any]:
        for old_field, new_field in PlanningAreas._FIELD_MAP.items():
            if new_field not in d:
                d[new_field] = d.pop(old_field)
        return d

    def get_nearest_to(self, point: GeoPt) -> Tuple[Optional[PlanningArea], float]:
        return (self.interval_tree.find_shape(point), 0)
        


from __future__ import annotations
from typing import Any, Generic, List, Optional, Tuple, TypeVar

from .comparable import Comparable

T = TypeVar("T", bound=Comparable)

class Node(Generic[T]):
    """
    Encapsulates a Node of the AVLTree.
    
    Fields:
        key (U): the value to be assigned to the node.
        left (Optional[Node[T]]): the left child of the node.
        right (Optional[Node[T]]): the right child of the node.
        height (int): the height of the tree rooted at the node.
        count (int): duplicate values are captured here, so we can have nodes with >1 counts.
        weight (int): the number of nodes on the tree rooted at the node.
    """
    key:    T
    left:   Optional[Node[T]]
    right:  Optional[Node[T]]
    height: int
    count:  int
    weight: int

    def __init__(self, key: T):
        self.key = key
        self.left = None
        self.right = None
        self.height = 1
        self.count = 1
        self.weight = 1

class AVLTree(Generic[T]):
    """
    Encapsulates an AVLTree capable of performing rotations, insertions and deletions.

    Fields:
        root (Optional[Node[T]]): the root of the tree.
    """
    root: Optional[Node[T]]
            
    def __init__(self):
        self.root = None
    
    def insert(self, key: T) -> None:
        """
        Inserts a key to the tree, triggering rotations.
        If the key already exists, then increment the key's count.
        Must be comparable or else the < > operators won't work.

        Args:
            key (T): key to be added.
        """
        def insert_helper(node: Optional[Node[T]], key: T) -> Optional[Node[T]]:
            if not node:
                return Node[T](key)
            elif key < node.key:
                node.left = insert_helper(node.left, key)
            elif key > node.key:
                node.right = insert_helper(node.right, key)
            else:
                node.count += 1
            self.set_height(node)
            self.set_weight(node)
            
            balance = self.get_balance(node)
            
            left_key = self.get_key(node.left)
            right_key = self.get_key(node.right)

            if balance > 1 and key < left_key:
                return self.right_rotate(node)
            if balance < -1 and right_key and key > right_key:
                return self.left_rotate(node)
            if balance > 1 and left_key and key > left_key:
                node.left = self.left_rotate(node.left)
                return self.right_rotate(node)
            if balance < -1 and key < right_key:
                node.right = self.right_rotate(node.right)
                return self.left_rotate(node)

            return node
        self.root = insert_helper(self.root, key)

    def delete(self, key: Optional[T]) -> None:
        """
        Deletes a key from the tree, triggering rotations.
        If the key has onle one count, then remove it from the tree.
        Otherwise, just decrement its count and update weights along the way.

        Args:
            key (T): key to be deleted.
        """
        def delete_helper(node: Optional[Node[T]], key: T) -> Optional[Node[T]]:
            if not node:
                return None
            elif key < node.key:
                node.left = delete_helper(node.left, key)
            elif key > node.key:
                node.right = delete_helper(node.right, key)
            elif node.count > 1:
                node.count -= 1
            else:
                if node.left is None:
                    temp = node.right
                    node = None
                    return temp
                elif node.right is None:
                    temp = node.left
                    node = None
                    return temp
                temp = self.get_min_value_node(node.right)
                node.key = temp.key
                node.count = temp.count
                node.right = delete_helper(node.right, temp.key)
            if not node:
                return node
            self.set_height(node)
            self.set_weight(node)
            balance = self.get_balance(node)

            if balance > 1 and self.get_balance(node.left) >= 0:
                return self.right_rotate(node)
            if balance < -1 and self.get_balance(node.right) <= 0:
                return self.left_rotate(node)
            if balance > 1 and self.get_balance(node.left) < 0:
                node.left = self.left_rotate(node.left)
                return self.right_rotate(node)
            if balance < -1 and self.get_balance(node.right) > 0:
                node.right = self.right_rotate(node.right)
                return self.left_rotate(node)
            return node
        if key is not None:
            self.root = delete_helper(self.root, key)
 
    def left_rotate(self, node: Optional[Node[T]]) -> Optional[Node[T]]:
        """
        Rotates the node left.

        Args:
            node (Optional[Node[T]]): node to be rotated.

        Returns:
            Optional[Node[T]]: the resulting root node.
        """
        if not node or not node.right:
            return node
        right = node.right
        right_left = right.left
        right.left = node
        node.right = right_left
 
        self.set_height(node)
        self.set_weight(node)
        self.set_height(right)
        self.set_weight(right)
        return right
 
    def right_rotate(self, node: Optional[Node[T]]) -> Optional[Node[T]]:
        """
        Rotates the node right.

        Args:
            node (Optional[Node[T]]): node to be rotated.

        Returns:
            Optional[Node[T]]: the resulting root node.
        """
        if not node or not node.left:
            return node
        left = node.left
        left_right = left.right
        left.right = node
        node.left = left_right
 
        self.set_height(node)
        self.set_weight(node)
        self.set_height(left)
        self.set_weight(left)
        return left
    
    def get_key(self, node: Optional[Node[T]]) -> Optional[T]:
        if not node:
            return None
        return node.key
 
    def get_height(self, node: Optional[Node[T]]) -> int:
        if not node:
            return 0
        return node.height
    
    def set_height(self, node: Node[T]) -> None:
        node.height = 1 + max(self.get_height(node.left), self.get_height(node.right))
        
    def get_weight(self, node: Optional[Node[T]]) -> int:
        if not node:
            return 0
        return node.weight
    
    def set_weight(self, node: Node[T]) -> None:
        node.weight = node.count + self.get_weight(node.left) + self.get_weight(node.right)
 
    def get_balance(self, node: Optional[Node[T]]) -> int:
        if not node:
            return 0
        return self.get_height(node.left) - self.get_height(node.right)
    
    def get_min_value(self) -> Optional[T]:
        """
        Gets the minimum value of the tree.

        Returns:
            T: the minimum value.
        """
        def get_min_value_helper(node: Optional[Node[T]]) -> Optional[T]:
            if not node:
                return None
            elif not node.left:
                return node.key
            return get_min_value_helper(node.left)
        return get_min_value_helper(self.root)
    
    def get_min_value_node(self, node: Node[T]) -> Node[T]:
        """
        Get the node containing the minimum value in the tree rooted at this node.

        Args:
            node (Node[T]): node to be queried.

        Returns:
            Node: node containing the minimum value.
        """
        if node is None or node.left is None:
            return node
        return self.get_min_value_node(node.left)

    def in_order(self) -> List[T]:
        """
        In order traversal of the tree done by recursion.

        Returns:
            List[T]: a list of keys in the tree in ascending order.
        """
        L = []
        def in_order_helper(node: Optional[Node[T]]) -> None:
            if not node:
                return
            in_order_helper(node.left)
            L.append(node.key)
            in_order_helper(node.right)
        in_order_helper(self.root)
        return L
    
    def pre_order(self) -> List[T]:
        """
        Pre order traversal of the tree done by recursion.

        Returns:
            List[T]: a list of keys in the tree in pre order.
        """
        L = []
        def pre_order_helper(node: Optional[Node[T]]) -> None:
            if not node:
                return
            L.append(node.key)
            pre_order_helper(node.left)
            pre_order_helper(node.right)
        pre_order_helper(self.root)
        return L
    
    def get_rank(self, key: T) -> int:
        """
        Gets the rank of a particular key in the tree.
        Does take into account duplicate values.

        Args:
            key (T): key to be queried.

        Returns:
            int: its rank, starting from 1.
        """
        def rank_helper(node: Optional[Node[T]], key: T) -> int:
            if not node:
                return 0
            elif key < node.key:
                return rank_helper(node.left, key)
            elif key > node.key:
                return self.get_weight(node) - self.get_weight(node.right) + rank_helper(node.right, key)
            return self.get_weight(node.left)
        return rank_helper(self.root, key) + 1
    


from __future__ import annotations
from os.path import dirname, join
from typing import Any, Callable, Dict, List

import pandas as pd

from .location import Location, Locations
from ..geom.elevation import ElevationMap

class BusStop(Location):
  """
  Object to encapsulate a Bus Stop in Singapore.

  Fields:
    services (List[str]): The bus services that go to this Bus Stop.
    road_name (str): The name of the road that this bus stop is on.
  """
  services: List[str]
  road_name: str

  def __init__(self, name: str, **kwargs):
    fields = ["lat", "lon", "shape", "services", "road_name"]
    self._try_setter(fields, kwargs)
    super().__init__(name, lat=self.lat, lon=self.lon, shape=self.shape)

class BusStops(Locations):
  _FIELD_MAP = {
    
  }

  def __init__(self, *bus_stops: BusStop, name: str="busstop"):
    super().__init__(*bus_stops, name=name)

  @staticmethod
  def get(blanks: bool=False, offline: bool=True) -> BusStops:
    raw_df = BusStops._get_data_handler(offline)
    data_dict = BusStops._get_data_cleaning(blanks)(raw_df)
    bus_stops = BusStops._get_data_compiling(data_dict)
    return BusStops(*bus_stops)

  @staticmethod
  def _get_data_handler(offline: bool) -> pd.DataFrame:
    if offline:
      return pd.read_csv(join(dirname(__file__), "assets/bus.csv"))
    print("Retrieving 'Bus' from Sheets...")
    raw_df = Locations.get_sheet("Bus")
    print("Retrieved.")
    return raw_df

  @staticmethod
  def _get_data_cleaning(blanks: bool=True) -> Callable[[pd.DataFrame], Dict[str, Any]]:
    def convert_into_dict(df: pd.DataFrame) -> Dict[str, Any]:
      buses_dict = {}
      for bus_stop, frame in df.groupby("BusStopCode"):
        buses_dict[bus_stop] = {
          "lat": frame.Latitude.to_list()[0],
          "lon": frame.Longitude.to_list()[0],
          "road_name": frame.RoadName.to_list()[0],
          "services": list(set(frame.ServiceNo.to_list()))
        }
      return buses_dict
    return convert_into_dict
      
  @staticmethod
  def _get_data_compiling(data_dict: Dict[str, Any]) -> List[BusStop]:
    bus_stops: List[BusStop] = []
    for name, info in data_dict.items():
      if ElevationMap.in_singapore(info["lat"], info["lon"]):
        bus_stops.append(BusStop(name, **BusStops._field_map(info)))
    return bus_stops

  @staticmethod
  def _field_map(d: Dict[str, Any]) -> Dict[str, Any]:
    for old_field, new_field in BusStops._FIELD_MAP.items():
      d[new_field] = d.pop(old_field)
    return d



from __future__ import annotations
from functools import cached_property
from typing import Any, Generic, List, Optional, Tuple, TypeVar

from shapely import geometry

from ..geom.geo_pt import GeoPt
from ..structures.bound import Bound
from ..structures.quick_sort import median_with_left_right

T = TypeVar("T")

class BoundsNode(Generic[T]):
    """
    Encapsulates a node of the BoundsTree.

    Fields:
        shape (geometry.polygon.Polygon): the shape to be represented by this node.
        value (T): any special value we want to associate with this shape.
        bound (Bound): the Bounds of this node's shape.
        big_bound (Bound): the Bounds that contains the shapes of this node as well as its descendants.
        level (str): whether we are comparing min_x, max_x, min_y, max_y.
        left (Optional[BoundsNode[T]]): the left child of the node.
        right (Optional[BoundsNode[T]]): the right child of the node.
    """
    shape:     geometry.polygon.Polygon
    value:     T
    bound:     Bound[GeoPt]
    big_bound: Bound[GeoPt]
    level:     str
    left:      Optional[BoundsNode[T]]
    right:     Optional[BoundsNode[T]]

    def __init__(self, shape: geometry.polygon.Polygon, value: T, level: str):
        """
        Initialiser for the BoundsNode[T] object.
        Left and right children are set to None as the node would be initialised as a leaf node.
        """
        self.shape = shape
        self.value = value
        self.bound = Bound[GeoPt].get_bound_from_shape(shape)
        self.big_bound = self.bound
        self.level = level
        self.left = None
        self.right = None

    @cached_property
    def next_level(self) -> str:
        """
        Based on the current level of the node, compute the next level.
        The cycle goes: min_x -> min_y -> max_x -> max_y.

        Returns:
            str: 'min_x' or 'min_y' or 'max_x' or 'max_y'
        """
        level_mapping = {"min_x": "min_y",
                         "min_y": "max_x",
                         "max_x": "max_y",
                         "max_y": "min_x",}
        return level_mapping[self.level]

    def add(self, shape: geometry.polygon.Polygon, value: T=None) -> None:
        """
        Adds a shape-value pair to the Node.
        It will create a new leaf if there is a vacancy in the right location.
        Otherwise, the arguments will be passed on to its children.

        Args:
            shape (geometry.polygon.Polygon): shape to be added.
            value (T, optional): value associated with the shape. Defaults to None.
        """
        bound: Bound = Bound.get_bound_from_shape(shape)
        self.big_bound.merge_with(bound)
        if getattr(bound, self.level) <= getattr(self.bound, self.level):
            if self.left == None:
                self.left = BoundsNode[T](shape, value, self.next_level)
            else:
                self.left.add(shape, value)
        else:
            if self.right == None:
                self.right = BoundsNode[T](shape, value, self.next_level)
            else:
                self.right.add(shape, value)

class BoundsTree(Generic[T]):
    """
    Encapsulates an BoundsTree containing many BoundsNode[T]s.
    Out of a set of polygons, interval tree seeks to optimise the search for
        which polygon contains a certain point.
    At each depth level, we compare alternating coordinates, starting with min_x.

    Fields:
        root (Optional[BoundsNode[T]]): the root of the tree.
        weight (int): the number of nodes in the tree.
        big_bound (Bound[U]): the Bound object containing all shapes in the tree.
    """
    root:      Optional[BoundsNode[T]]
    weight:    int
    big_bound: Bound[GeoPt]

    def __init__(self):
        """
        Initialiser for the BoundsTree object.
        Big bound is initialised to infinity.
        """
        self.root = None
        self.weight = 0
        self.big_bound = Bound[GeoPt](float("inf"), float("-inf"), float("inf"), float("-inf"))

    def __str__(self) -> str:
        return str(self.in_order())

    def __repr__(self) -> str:
        return f"<BoundsTree: {self.big_bound}>"

    def add_all(self, shapes: List[geometry.polygon.Polygon], values: List[T]) -> None:
        """
        Adds all shape-value pairs to the tree, creating a new root if it does not exist.

        Args:
            shapes (List[geometry.polygon.Polygon]): list of shapes to be added.
            values (List[T]): values associated with each shape.
        """
        zipped = list(zip(shapes, values))
        sorted_zip: List[Tuple[geometry.polygon.Polygon, T]] = []
        
        def append_median(_list: List[Tuple[geometry.polygon.Polygon, T]], xymm: str) -> None:
            if len(_list) == 0:
                return
            left, mid, right = median_with_left_right(_list,
                                                      comparator=lambda item: getattr(Bound.get_bound_from_shape(item[0]), xymm))
            if mid is not None:
                sorted_zip.append(mid)
            if xymm == "min_x":
                append_median(left, "min_y")
                append_median(right, "min_y")
            elif xymm == "min_y":
                append_median(left, "max_x")
                append_median(right, "max_x")
            elif xymm == "max_x":
                append_median(left, "max_y")
                append_median(right, "max_y")
            else:
                append_median(left, "max_x")
                append_median(right, "max_x")
        append_median(zipped, "min_x")
        
        if self.root == None:
            self.root = BoundsNode[T](sorted_zip[0][0], sorted_zip[0][1], "min_x")
            self.big_bound = Bound.get_bound_from_shape(shapes[0])
            sorted_zip = sorted_zip[1:]
            self.weight += 1
            
        for shape, value in sorted_zip:
            self.root.add(shape, value)
            self.big_bound.merge_with(Bound.get_bound_from_shape(shape))
            self.weight += 1

    def find_bounds_containing(self, point: GeoPt) -> List[Bound]:
        """
        Recursively finds all the bounds objects that contain the point.

        Args:
            point (T): point to be queried.

        Returns:
            List[Bound]: Bound objects that contain the point.
        """
        L: List[Bound] = []
        def find_bounds_containing_helper(node: Optional[BoundsNode[T]]) -> None:
            if not node or not node.big_bound.contains(point):
                return
            L.append(node.bound)
            find_bounds_containing_helper(node.left)
            find_bounds_containing_helper(node.right)
        find_bounds_containing_helper(self.root)
        return L

    def find_shape(self, point: GeoPt) -> Optional[T]:
        """
        Finds the singular shape in the tree that contains the point and returns its value.

        Args:
            point (T): point to be queried.

        Returns:
            T: value associated with the shape queried.
        """
        value: List[Optional[T]] = [None]
        def find_shape_helper(node: Optional[BoundsNode[T]]) -> None:
            if not node or not node.big_bound.contains(point):
                return
            if not node.shape.contains(point):
                find_shape_helper(node.left)
                find_shape_helper(node.right)
                return
            value[0] = node.value
        find_shape_helper(self.root)
        return value[0]


    def in_order(self) -> List[Bound]:
        """
        In order traversal of the tree done by recursion.

        Returns:
            List[Bound]: a list of bounds in the tree.
        """
        L: List[Bound] = []
        def in_order_helper(node: Optional[BoundsNode[T]]) -> None:
            if not node:
                return
            in_order_helper(node.left)
            L.append(node.bound)
            in_order_helper(node.right)
        in_order_helper(self.root)
        return L

    def pre_order(self) -> List[Bound]:
        """
        Pre order traversal of the tree done by recursion.

        Returns:
            List[Bound]: a list of bounds in the tree.
        """
        L: List[Bound] = []
        def pre_order_helper(node: Optional[BoundsNode[T]]) -> None:
            if not node:
                return
            L.append(node.bound)
            pre_order_helper(node.left)
            pre_order_helper(node.right)
        pre_order_helper(self.root)
        return L
    


from __future__ import annotations
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, Optional, Set, Tuple
import re

from .mrt_line import MRTLine, get_mrt_lines

class PlatformCode:
    code: str
    
    _lower_terminuses: Set[str] = {"EW1", "CG0",
                                   "NS1", "BP1",
                                   "NE1", "PW0", "PE0", "SE0", "SW0",
                                   "CC1", "CE0",
                                   "DT0",
                                   "TE1",
                                   "JE0", "JS1", "JW0",
                                   "CR1", "CP1",}
    _upper_terminuses: Set[str] = {"EW33", "CG2",
                                   "NS28", "BP14",
                                   "NE18", "PW8", "PE8", "SE6", "SW9",
                                   "CC35", "CE2",
                                   "DT37",
                                   "TE31",
                                   "JE7", "JS12", "JW5",
                                   "CR13", "CP4",}
    _exceptions: Dict[str, Tuple[str, str]] = {"NS3": ("NS2", "NS3A"),
                                               "NS4": ("NS3A", "NS5"),
                                               "NS3A": ("NS3", "NS4"),
                                               "TE22": ("TE21", "TE22A"),
                                               "TE23": ("TE22A", "TE24"),
                                               "TE22A": ("TE22", "TE23"),}
    _line_mapping: Dict[str, str] = {"CP": "CR",
                                     "CE": "CC",
                                     "CG": "EW",}
    
    MRT_LINES = get_mrt_lines()
    
    def __init__(self, code: str):
        self.code = code
    
    def __eq__(self, other: PlatformCode) -> bool:
        return self.code == other.code
    
    def __lt__(self, other: PlatformCode) -> bool:
        return self.number < other.number
    
    def __hash__(self) -> int:
        return hash(self.code)
    
    def __str__(self) -> str:
        return f"<PlatformCode: {self.code}>"

    def __repr__(self) -> str:
        return f"<PlatformCode: {self.code}>"
    
    @cached_property
    def prefix(self) -> str:
        return re.findall(r"[A-Z]+", self.code)[0]
    
    @cached_property
    def number(self) -> int:
        regex_search_result = re.findall(r"\d+", self.code)
        if len(regex_search_result) == 0:
            return 0
        return int(regex_search_result[0])
    
    @cached_property
    def suffix(self) -> Optional[str]:
        regex_search_result = re.findall(r"[A-Z]+", self.code)
        if len(regex_search_result) < 2:
            return None
        return regex_search_result[1]
    
    @cached_property
    def is_lower_terminus(self) -> bool:
        return self.code in PlatformCode._lower_terminuses
    
    @cached_property
    def is_higher_terminus(self) -> bool:
        return self.code in PlatformCode._upper_terminuses
    
    @cached_property
    def next_code(self) -> Optional[PlatformCode]:
        if self.is_higher_terminus:
            return None
        if self.code in PlatformCode._exceptions:
            return PlatformCode(PlatformCode._exceptions[self.code][1])
        return PlatformCode(self.prefix + str(self.number + 1))
    
    @cached_property
    def prev_code(self) -> Optional[PlatformCode]:
        if self.is_lower_terminus:
            return None
        if self.code in PlatformCode._exceptions:
            return PlatformCode(PlatformCode._exceptions[self.code][0])
        return PlatformCode(self.prefix + str(self.number - 1))
        
    @cached_property
    def line(self) -> MRTLine:
        if self.prefix in PlatformCode._line_mapping:
            new_prefix = PlatformCode._line_mapping[self.prefix]
            line = PlatformCode.MRT_LINES[new_prefix]
            if line is not None:
                return line
        line = PlatformCode.MRT_LINES[self.prefix]
        if line is not None:
            return line
        raise ValueError(self.prefix)


from __future__ import annotations
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
import json

from gsheets import models, Sheets
from tqdm import tqdm
# import contextily as ctx
import geopandas as gpd
import numpy as np
import pandas as pd
import requests
import shapely.geometry

from ..error.value_error.out_of_bounds_error import OutOfBoundsError
from ..geom.elevation import ElevationMap
from ..geom.geo_pt import GeoPt
from ..geom.shape import Shape
from ..structures.bound import Bound
from ..structures.kdtree import KDTree
from ..utils.float import is_float

class Location(GeoPt, ABC):
  """
  Object to encapsulate a location in Singapore.
  
  Fields:
    name (str): the name of the location.
    lat (Optional[float]): the latitude coordinate of the location.
    lon (Optional[float]): the longitude coordinate of the location.
    shape (Optional[Shape]): the shape representing the location, if available.
    _nearest (Dict[str, Tuple[Location, float]]): the dictionary storing
      the nearest (malls, schools etc.) to that particular location.
  """
  name:   str
  lat:    Optional[float]
  lon:    Optional[float]
  shape:  Optional[Shape] = None
  _nearest: Dict[str, Tuple[Optional[GeoPt], float]]
  _within: Dict[str, List[GeoPt]]

  def __init__(self, name: str, lat: Optional[float]=None, lon: Optional[float]=None, shape: Optional[Shape]=None):
    self.name = name
    self.shape = shape
    lat, lon = self._get_coords(lat, lon)
    self._nearest = {}
    self._within = {}
    if lat and lon:
      super().__init__(lat=lat, lon=lon)
    else:
      super().__init__(lat=float("inf"), lon=float("inf"))

  def __str__(self) -> str:
    type_ = str(type(self)).split(".")[-1].split("'")[0]
    return f"<{type_}: {self.name}, ({self.lat}, {self.lon})>"

  def __repr__(self) -> str:
    type_ = str(type(self)).split(".")[-1].split("'")[0]
    return f"<{type_}: {self.name}>"
  
  def _get_coords(self, lat: Optional[float], lon: Optional[float]) -> Tuple[Optional[float], Optional[float]]:
    """
    If latitude or longitude are not available, first check whether the shape is available.
    If shape is available, then compute the center of the shape.
    Otherwise, query developers.onemap.sg to obtain the lat long based on its name.

    Args:
      lat (Optional[float]): latitude.
      lon (Optional[float]): longitude.

    Returns:
      Tuple[Optional[float], Optional[float]]: new lat long values.
    """
    if lat and lon and is_float(lat) and is_float(lon):
      return lat, lon
    if self.shape is not None:
      return GeoPt.from_bound(Bound.get_bound_from_shape(self.shape)).coords_as_tuple_latlong()
    try:
      return self._get_coords_from_api(self.name)
    except IndexError:
      print(self.name, "could not be found.")
      return None, None

  def _get_coords_from_api(self, name: str) -> Tuple[float, float]:
    """
    Queries developers.onemap.sg to obtain lat long values.

    Args:
      name (str): name of the location to be queried.

    Returns:
      Tuple[float, float]: newly mapped lat long values.
    """
    search_term = "+".join(name.split())
    search_url = ("https://developers.onemap.sg/commonapi"
            f"/search?searchVal={search_term}"
            "&returnGeom=Y&getAddrDetails=Y")
    coords = json.loads(requests.get(search_url).text)["results"][0]

    return float(coords["LATITUDE"]), float(coords["LONGITUDE"])
  
  def _try_setter(self, fields: List[str], items: Dict[str, Any]) -> None:
    """
    For each field in the specified fields list,
      the method will try to assign it to the object.
    If not possible, then it will just skip it.

    Args:
      fields (List[str]): list of fields to be included in the object.
      items (Dict[str, Any]): dictionary containing the fields and values.
        May not always contain all the fields needed.
    """
    for field in fields:
      if field in items and items[field] is not None:
        setattr(self, field, items[field])
      else:
        setattr(self, field, None)
  
  @cached_property
  def elevation(self) -> Optional[float]:
    """
    Lazily gets elevation value of the location.

    Returns:
      Optional[float]: elevation at the location.
    """
    if self.lat and self.lon:
      return ElevationMap.get_elevation(lat=self.lat, lon=self.lon)
    return None

  def get_distance(self, point: Location, base: bool=True) -> float:
    """
    Gets distance from the location to another point on Earth.

    Args:
      point (Location): the destination point to query.

    Returns:
      float: distance in km.
    """
    if self.shape is not None:
      if self.shape.contains(point):
        return 0
      if base:
        if point.shape and self.shape.exterior and point.shape.exterior:
          distance = (self.shape.exterior.distance(point) * 111.33
                + point.shape.exterior.distance(self) * 111.33
                - super().get_distance(point))
          return max(round(distance, 4), 0)
      return self.shape.get_nearest(point)[1]
    return super().get_distance(point)

  def map_nearby(self, locations: Locations, threshold: float, name: str="") -> List[GeoPt]:
    """
    Gets all the locations within a certain radius around self.
    
    Args:
      locations (Locations): target locations to fall within the circle.
      threshold (float): maximum distance permitted for the circle.
      name (str): name to assign to this group of locations (malls etc.)
      
    Returns:
      List[GeoPt]: points that fall within the range.
    """
    within: List[GeoPt] = []
    for location in locations:
      if self.get_distance(location) <= threshold:
        within.append(location)
    self._within[name] = within
    setattr(self, "nearby_"+name, within)
    return within

  def nearest(self, name: str) -> Tuple[Optional[GeoPt], float]:
    """
    If already generated, returns the nearest location in that group,
      in the form of a point-distance pair.
    Else, warns the user.

    Args:
      name (str): group of locations to be queried.

    Returns:
      Tuple[Optional[GeoPt], float]: point-distance pair.
    """
    if name not in self._nearest:
      joined_keys = ', '.join([*self._nearest.keys()])
      print(f"Nearest '{name}' not defined. Available options: {joined_keys}.")
      return (None, float("inf"))
    return self._nearest[name]

  def nearest_distance(self, name: str) -> float:
    """
    Gets the distance from the point-distance pair.

    Args:
      name (str): group of locations to be queried.

    Returns:
      float: distance to nearest point to location.
    """
    return self.nearest(name)[1]

  def nearest_location(self, name: str) -> Optional[GeoPt]:
    """
    Gets the point from the point-distance pair.

    Args:
      name (str): group of locations to be queried.

    Returns:
      Optional[GeoPt]: nearest point to location.
    """
    return self.nearest(name)[0]

  def map_nearest(self,
          locations: Locations,
          name: str="") -> Tuple[Optional[GeoPt], float]:
    """
    From a set of locations, get the nearest one to this location.

    Args:
      name (str): name to assign to this group of locations (malls etc.)
      locations (Locations): collection of locations to search from.

    Returns:
      Tuple[GeoPt, float]: pair of point and distance to point.
    """
    name = name if name != "" else locations.name
    nearest_point, nearest_distance = locations.get_nearest_to(self)
    self._nearest[name] = (nearest_point, nearest_distance) if nearest_point else (None, float("inf"))
    setattr(self, name, self._nearest[name])
    setattr(self, "nearest_"+name, self._nearest[name][0])
    setattr(self, "distance_to_"+name, self._nearest[name][1])
    return self._nearest[name]

class Locations(ABC):
  """
  Object encapsulating a collection of locations.
  Facilitates the searching of nearest points (using KDTree).
  Enables filtering, sorting regex searching of locations.

  Fields:
    kdtree (KDTree[Dict]): KDTree representing the locations.
      This reduces the expected time complexity of searching nearest points to O(logN).
    locations (Dict[str, Location]): dictionary storing the locations of the group,
      indexed by name to make it easier to access them.
    name (str): the name assigned to this group of locations.
  """
  kdtree:  KDTree[GeoPt]
  locations: Dict[str, Location]
  name:    str

  _SHEET_ID = "1M9Ujc54yZZPlxOX3yxWuqcuJOxzIrDYz4TAFx8ifB8c"

  def __init__(self, *locations: Location, name: str):
    """
    Initialiser for the Locations object.

    Args:
      name (str): name to assign to the group of locations.
    """
    self.name = name
    self.kdtree = KDTree[GeoPt]()
    self.locations = {}
    self.kdtree.add_all(*locations)
    for location in locations:
      if location.lat == float("inf"):
        continue
      if location.name not in self.locations:
        self.locations[location.name] = location
      else:
        self.locations[location.name+" "+type(location).__name__] = location

  def __getitem__(self, search_term="") -> Optional[Location]:
    """
    Searches for a particular location in the group, using relaxed searching.
    If only one result exists, then return it.
    Elif more than one result exists, warn the user and return the first result.
    Else return None.

    Args:
      search_term (str, optional): the string to look for
        in the names of the locations. Defaults to "".

    Returns:
      Optional[Location]: a single location that matches the criteria.
    """
    if search_term in self.locations:
      return self.locations[search_term]
    search_results = dict(filter(lambda location: search_term in location[0],
                   self.locations.items()))
    if len(search_results) == 1:
      return list(search_results.values())[0]
    if len(search_results) == 0:
      print(f"{search_term} not found. Please try again.")
      return None
    print(f"'{search_term}' produced multiple locations "
        f"({', '.join(list(search_results.keys()))}). "
        "Returning first result.")
    return list(search_results.values())[0]

  def __iter__(self) -> LocationIterator:
    return LocationIterator(self.locations)
  
  def insert(self,
        to_insert: List[Location],
        name: str="") -> Locations:
    """
    Adds locations to the existing set of locations.
    Returns another Locations object containing these new locations.
    Inserts can be chained together like in locations.insert(i1).insert(i2)
    For example, [L1, L2].

    Args:
      to_insert (List[Location]): list of locations to add to the existing set.
      name (str, optional): optional name to be given to this subset of locations. Defaults to "".

    Returns:
      Locations: superset of the original locations.
    """
    name_to_use = self.name if name == "" else name
    appended_locations = list(self.locations.values()) + to_insert
    locations = type(self)(*appended_locations, name=name_to_use)
    return locations

  @staticmethod
  @abstractmethod
  def get(blanks: bool, offline: bool) -> Locations:
    """
    The full process for obtaining data from a dataset.
    This may take varying amounts of time depending on the type of Location being processed.

    Args:
      blanks (bool): whether to allow blank fields in the final set.
      offline (bool): whether to process the offline (static) dataset,
        or the online (dynamic) dataset. The dynamic dataset will take a
        longer time to process.

    Returns:
      Locations: the group of locations after processing the data.
    """
    pass

  @staticmethod
  @abstractmethod
  def _get_data_handler(offline: bool) -> pd.DataFrame:
    """
    Process for extracting data from the dataset.

    Args:
      offline (bool): whether to process the static or dynamic dataset.

    Returns:
      pd.DataFrame: the dataframe containing the relevant information.
    """
    pass

  @staticmethod
  @abstractmethod
  def _get_data_cleaning(blanks: bool) -> Dict[str, Any]:
    """
    Process for cleaning data extracted from the dataset.

    Args:
      blanks (bool): whether to accept any blank fields.

    Returns:
      Dict[str, Any]: dictionary storing the name of the location,
        mapped to the supplementary information.
    """
    pass

  @staticmethod
  @abstractmethod
  def _get_data_compiling(data_dict: Dict) -> Locations:
    """
    Creates the Location objects based on the data processed earlier.

    Args:
      data_dict (Dict): data dictionary to be processed.

    Returns:
      Locations: group of locations in the final set.
    """
    pass
  
  @cached_property
  def count(self):
    return len(self.locations.keys())
  
  def filter(self,
         location_filter: Callable[[Location], bool]=lambda location: True,
         name: str="") -> Locations:
    """
    Filters out locations based on certain criteria (lambda function).
    Returns another Locations object containing these new locations.
    Filters can be chained together like in locations.filter(f1).filter(f2)
    For example, lambda location: location.lat > 1.38.

    Args:
      location_filter (Callable, optional): lambda function to determine whether a locations passes the filter.
        Defaults to lambda location:True.
      name (str, optional): optional name to be given to this subset of locations. Defaults to "".

    Returns:
      Locations: subset of the original locations.
    """
    name_to_use = self.name if name == "" else name
    filtered_locations = list(filter(location_filter, self.locations.values()))
    locations = type(self)(*filtered_locations, name=name_to_use)
    return locations
  
  def get_by_regex(self, search_term: str="") -> Dict[str, Location]:
    """
    Gets locations whose names match the regex pattern.

    Args:
      search_term (str, optional): regex pattern to search for. Defaults to "".

    Returns:
      Dict[str, Location]: filtered dictionary based on the pattern matching.
    """
    search_results = dict(filter(lambda location: search_term in location[0],
                   self.locations.items()))
    return search_results

  def get_nearest_to(self, point: GeoPt) -> Tuple[Optional[GeoPt], float]:
    """
    For an external point, find which of this group's locations is the closest.

    Args:
      point (GeoPt): external point to query.

    Returns:
      Tuple[GeoPt, float]: point-distance pair.
    """
    nearest_point, distance = self.kdtree.nearest(point)
    return (nearest_point, distance) if nearest_point else (None, float("inf"))
  
  @staticmethod
  def get_sheet(name: str) -> pd.DataFrame:
    """
    Extracts data from Google Sheets, given the name of the sheet.

    Args:
      name (str): name of the sheet to be downloaded.

    Returns:
      pd.DataFrame: data extracted from Google Sheets.
    """
    sheets: Union[models.SpreadSheet, List[object]] = Sheets.from_files('client_secrets.json','~/storage.json')[Locations._SHEET_ID]
    return sheets.find(name).to_frame() if isinstance(sheets, models.SpreadSheet) else pd.DataFrame()

  def group_by(self,
         comparator: Callable[[Location], str]=lambda location: location.name) -> Dict[str, Locations]:
    locations_dict: Dict[str, List[Location]] = {}
    for _, location in self.locations.items():
      to_compare = comparator(location)
      if to_compare not in locations_dict:
        locations_dict[to_compare] = []
      locations_dict[to_compare].append(location)
    
    return {k: type(self)(*v, name=k) for k, v in locations_dict.items()}
      
  def map_nearby_to(self,
										locations: Locations,
										threshold: float,
										locations_name: str="",
										progress_bar: bool=False) -> Dict[str, List[GeoPt]]:
    """
    For each location in this set, apply the nearby method to each one of an external location.
    For example, malls.map_nearby_to(schools) will map each mall to schools that are near to the mall.

    Args:
      locations (Locations): group of destination locations to map to.
      locations_name (str): what to name the group of locations.
      threshold (float): maximum distance permitted to quality for this set.

    Returns:
      List[Tuple[str, Tuple[Location, float]]]: the result of the mapping.
    """
    to_return: Dict[str, List[GeoPt]] = {}
    iterable = tqdm(self.locations.items()) if progress_bar else self.locations.items()
    for name, location in iterable:
      locations_name = locations.name if locations.name != "" else locations.name
      nearby = location.map_nearby(locations, threshold, locations_name)
      to_return[name] = nearby
    return to_return    
      
  def map_nearest_to(self,
             				 locations: Locations,
                     locations_name: str="",
                     progress_bar: bool=False,
                     dist: bool=True) -> Dict[str, Tuple[Location, float]]:
    """
    For each location in this set, apply the nearest method to each one of an external location.
    For example, malls.map_nearest_to(schools) will map each mall to its nearest school.

    Args:
      locations (Locations): group of destination locations to map to.
      locations_name (str): what to name the group of locations.
      prefix (str): what to put in front of the name.

    Returns:
      List[Tuple[str, Tuple[Location, float]]]: the result of the mapping.
    """
    to_return = {}
    iterable = tqdm(self.locations.items()) if progress_bar else self.locations.items()
    for name, location in iterable:
      locations_name = locations.name if locations.name != "" else locations.name
      location.map_nearest(locations, locations_name)
      to_return[name] = location.nearest(locations_name)
    return to_return
   
  # def plot(self, zoom: int=13, figsize=(20, 20), alpha=0.4, color="red") -> None:
  #   if not 12 <= zoom <= 19:
  #     raise OutOfBoundsError(zoom, 12, 19)
  #   gdf = gpd.GeoDataFrame([location.shape if hasattr(location, "shape") else location for location in self.locations.values()], columns=["geometry"])
  #   ax = gdf.geometry.plot(figsize=figsize, alpha=alpha, color=color)
  #   ctx.add_basemap(ax, zoom=zoom, crs="EPSG:4326", source=ctx.providers.OneMapSG.Night)

  def show(self, attr: str="name") -> Dict[str, Any]:
    """
    Shows a particular attribute of all the locations.

    Args:
      attr (str, optional): the attribute to show. Defaults to "name".

    Returns:
      Dict[str, Any]: name-attribute pair.
    """
    return {name: getattr(value, attr) for name, value in self.locations.items()}
  
  def show_by_lambda(self,
             getter: Callable[[Location], Any]=lambda location: location.name) -> Dict[str, Any]:
    """
    Instead of showing a particular attribute, this method allows the user to set up custom show functions.
    For example, lambda location: location.lat + location.lon.

    Args:
      getter (Callable[[Location], Any], optional): custom show function. Defaults to lambda location:location.name.

    Returns:
      Dict[str, Any]: name-value pair with the values being defined in the custom show function.
    """
    return {name: getter(value) for name, value in self.locations.items()}

  def sort(self,
       attr: str="name",
       show: str="",
       reverse=False) -> Dict[str, Location]:
    """
    Sorts the locations based on a particular attribute, or custom function.
    For example, lambda location: location.lat will sort them by latitude.

    Args:
      attr: the attribute to sort by. Defaults to name.
      show: the attribute to show in the end result. Defaults to be the same as attr.
      reverse (bool, optional): whether to reverse the values. Defaults to False.

    Returns:
      Dict[str, Location]: name-value pair with the values being defined in the custom show function.
    """
    if show == "":
      show = attr
    return {k: getattr(v, show) for k, v in sorted(self.locations.items(),
                             key=lambda pair: getattr(pair[1], attr),
                             reverse=reverse)}
  
  def sort_by_lambda(self,
             comparator: Callable[[Location], Any]=lambda location: location.name,
             show: Optional[Callable[[Location], Any]]=None,
             reverse=False) -> Dict[str, Location]:
    """
    Sorts the locations based on a particular attribute, or custom function.
    For example, lambda location: location.lat will sort them by latitude.

    Args:
      comparator (Callable[[Location], Any], optional): the lambda function used to compare elements.
        Defaults to lambda location: location.name.
      show (Callable[[Location], Any], optional): the lambda function used to show information contained
        within the location. Defaults to be the same as the comparator.
      reverse (bool, optional): whether to reverse the values. Defaults to False.

    Returns:
      Dict[str, Location]: name-value pair with the values being defined in the custom show function.
    """
    if not show:
      show = comparator
    return {k: show(v) for k, v in sorted(self.locations.items(),
                        key=lambda item: comparator(item[1]),
                        reverse=reverse)}

  def to_dict(self, subset: List[str]=[]) -> Dict[str, Dict[str, Any]]:
    return {
      k: {
        key: getattr(v, key) for key in subset
      } for k, v in self.locations.items()
    }

  def to_df(self, subset: List[str]=[]) -> pd.DataFrame:
    return pd.DataFrame.from_dict(self.to_dict(subset)).T

class LocationIterator:
  """
  Iterator object for the locations in Locations.
  """
  def __init__(self, locations: Dict[str, Location]):
    self._locations = list(locations.items())
    self._index = 0
    self._len = len(locations)
  
  def __next__(self) -> Location:
    if self._index < self._len:
      to_return = self._locations[self._index][1]
      self._index += 1
      return to_return
    raise StopIteration


from .avltree import AVLTree
from typing import Any, Callable, Dict, Generic, List, Optional, Set, Tuple, TypeVar

from .comparable import Comparable

T = TypeVar("T", bound=Comparable)
U = TypeVar("U")

class PriorityQueue(Generic[U, T]):
    """
    Encapsulates a PriorityQueue to enq and deq items in O(logN) time.

    Fields:
        tree (AVLTree[T]): the tree to keep track of the items.
    """
    tree: AVLTree[T]
    keys: Dict[T, List[U]]
    reverse_keys: Dict[U, T]
    comparator: Callable[[U], T]
    
    def __init__(self, comparator: Callable[[U], T]):
        self.tree = AVLTree[T]()
        self.keys = {}
        self.reverse_keys = {}
        self.comparator = comparator
        
    def is_empty(self) -> bool:
        return self.tree.root is None
    
    def enq(self, value: U) -> None:
        key = self.comparator(value)
        if key not in self.keys:
            self.keys[key] = [value]
        else:
            self.keys[key].append(value)
        self.tree.insert(key)
        self.reverse_keys[value] = key
    
    def deq(self) -> Optional[U]:
        key = self.tree.get_min_value()
        if key is None:
            return
        self.tree.delete(key)
        return self.keys[key].pop()
    
    def update(self, value: U) -> None:
        key = self.comparator(value)
        if value in self.reverse_keys:
            old_key = self.reverse_keys[value]
            self.keys[old_key].remove(value)
            self.reverse_keys[value] = key
            self.tree.delete(key)
        self.enq(value)
        
            


from __future__ import annotations
from typing import List

from .mrt_line import MRTLine
from ...geom.geo_pt import GeoPt
from ...geom.line import Line
from . import platform

class Connection(Line):
    lower: platform.Platform
    upper: platform.Platform
    line: MRTLine
    
    def __init__(self,
                 lower: platform.Platform,
                 upper: platform.Platform,
                 line: MRTLine,
                 segment: List[GeoPt]):
        super().__init__(segment)
        self.lower = lower
        self.upper = upper
        self.line = line
    
    # def split_on(self, mid: locations.mrt.platform.Platform) -> None:
    #     lower_connection = Connection(self.lower, mid, self.line)
    #     upper_connection = Connection(mid, self.upper, self.line)
    #     self.lower.upper_connection = lower_connection
    #     self.upper.lower_connection = upper_connection
    #     mid.lower_connection = lower_connection
    #     mid.upper_connection = upper_connection


from setuptools import setup, find_packages


setup(
  name='halfgeo',
  version='1.0.12',
  license='MIT',
  author="Izz Hafeez",
  author_email='izzhafeez@gmail.com',
  packages=find_packages('src'),
  package_dir={'': 'src'},
  url='https://github.com/mynameizzhafeez/geo',
  keywords='geospatial singapore',
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
  include_package_data=True,
  install_requires=[
      'pandas',
      'geopandas',
      'numpy',
      'shapely',
      'gsheets',
      'Pillow'
    ],

)


from math import atan2, cos, radians, sin, sqrt

from shapely import geometry

class DistanceCalculator:
    """
    Outputs the geographic distance between two points on Earth,
        considering its curvature
    Methods:
        get_distance_between: outputs the distance (km) between two
            shapely.geometry.Point objects
        get_distance_between_xy: outputs the distance (km) between two
            points, represented as two sets of latlongs
    """
    @staticmethod
    def get_distance(p1: geometry.Point,
                     p2: geometry.Point) -> float:
        """
        Outputs the distance (km) between two shapely.geometry.Point objects.
        Will perform different calculations based on whether they are Pts or GeoPts.

        Args:
            p1 (geometry.Point): starting point.
            p2 (geometry.Point): ending point.

        Raises:
            ValueError: GeoPts cannot be calculated with Pts and vice versa.

        Returns:
            float: distance, either in km or units.
        """
        if (not isinstance(p1, geometry.Point)
            or not isinstance(p2, geometry.Point)):
            raise ValueError("p1 and p2 must both be of type shapely.geometry.Point")
        return DistanceCalculator.get_distance_xy(p1.y, p1.x, p2.y, p2.x)

    @staticmethod
    def get_distance_xy(lat1: float,
                        lon1: float,
                        lat2: float,
                        lon2: float) -> float:
        """
        Outputs the distance (km) between two
            points, represented as two sets of latlongs.

        Args:
            lat1 (float): starting latitude.
            lon1 (float): starting longitude.
            lat2 (float): ending latitude.
            lon2 (float): ending longitude.

        Returns:
            float: distance in km between the two points.
        """
        R = 6371
        lat1 = radians(lat1)
        lon1 = radians(lon1)
        lat2 = radians(lat2)
        lon2 = radians(lon2)
        dlat = lat2 - lat1
        dlon = lon2 - lon1
        a = sin(dlat/2)**2 + cos(lat1)*cos(lat2)*sin(dlon/2)**2
        return round(R*2*atan2(sqrt(a), sqrt(1-a)), 4)
    
    @staticmethod
    def get_distance_basic(p1: geometry.Point, p2: geometry.Point) -> float:
        return 111.33*((p2.y-p1.y)**2+(p2.x-p1.x)**2)**0.5

    @staticmethod
    def get_distance_xy_basic(lat1: float,
                              lon1: float,
                              lat2: float,
                              lon2: float) -> float:
        return 111.33*((lat2-lat1)**2+(lon2-lon1)**2)**0.5




from __future__ import annotations
from functools import cached_property
from typing import List, Generic, get_args, Optional, Tuple, Type, TypeVar

from shapely import geometry
from shapely.ops import nearest_points

from ..geom.geo_pt import GeoPt
from ..geom.pointable import Pointable
from ..structures.bound import Bound
from ..structures.kdtree import KDTree

T = TypeVar("T", bound=Pointable)

class Line(geometry.LineString, Generic[T]):
    """
    This class encapsulates a Line object, with points represented
        as a KDTree to facilitate the finding of nearest points.

    Fields:
        points (KDTree): a KDTree full of points representing the Line.
    """
    points: KDTree[T]

    def __init__(self, points: List[T]):
        """
        Initialiser for the Line object.
        Creates an empty KDTree to store the points.

        Args:
            points (list): ordered list of points to be included in the line.
        """
        super().__init__(points)
        self.points = KDTree()
        self.points.add_all(*points)
        
    @cached_property
    def length(self) -> float:
        """
        Calculates the total length of the line.

        Returns:
            float: total length in metres.
        """
        dist = 0
        for i in range(len(self.coords)-1):
            lower = GeoPt(self.coords[i][1], self.coords[i][0])
            higher = GeoPt(self.coords[i+1][1], self.coords[i+1][0])
            dist += lower.get_distance(higher)
        return dist
        
    @staticmethod
    def from_linestring(line: geometry.LineString) -> Line[T]:
        # return Line[T]([get_args(T)[0](point.x, point.y) for point in line.coords])
        return Line[T]([GeoPt(point[1], point[0]) for point in line.coords])

    def get_nearest(self, point: T) -> Tuple[Optional[T], float]:
        """
        Gets the nearest point on the line to the point queried. Returns a point in the middle of the line.

        Args:
            point (T): the target point.

        Returns:
            Tuple[Optional[T], float]: the point-distance pair
                for the closest point from the line to the target.
        """
        nearest_point = nearest_points(self, point)[0].coords[0]
        moved_point: T = point.move_to(new_x=nearest_point[0], new_y=nearest_point[1])
        nearest_distance = moved_point.get_distance(point)
        return (moved_point, nearest_distance)
    
    def get_nearest_kd(self, point: T) -> Tuple[Optional[T], float]:
        """
        Gets the nearest point on the line to the point queried.

        Args:
            point (T): the target point.

        Returns:
            Tuple[Optional[T], float]: the point-distance pair
                for the closest point from the line to the target.
        """
        return self.points.nearest(point)

    def get_bounds(self) -> Bound:
        """
        Gets the bounds of the line (min-max values for x and y).

        Returns:
            Bound: bound object that surrounds the line.
        """
        return self.points.bound


from __future__ import annotations
from typing import List, Optional, Tuple

import shapely.geometry

from . import geo_pt
from .pointable import Pointable
from ..structures.bound import Bound

class Pt(shapely.geometry.Point, Pointable):
    """
    Object encapsulating a point, facilitating distance computation.
    
    Fields:
        x (float): x-coordinate for the point.
        y (float): y-coordinate for the point.
    """

    def __init__(self, x: float, y: float) -> None:
        """
        Initialiser for the Pt object.

        Args:
            x (float): x-coordinate for the point.
            y (float): y-coordinate for the point.
        """
        super().__init__(x, y)

    def __str__(self) -> str:
        return f"PT ({self.x}, {self.y})"
        
    def __repr__(self) -> str:
        return f"({self.x}, {self.y})"
    
    @property
    def x(self) -> float:
        return self.x
    
    @property
    def y(self) -> float:
        return self.y

    @staticmethod
    def from_bound(bound: Bound[Pt]) -> Pt:
        return Pt((bound.max_x+bound.min_x)/2, (bound.max_y+bound.min_y)/2)

    @staticmethod
    def from_point(point: shapely.geometry.Point) -> Pt:
        """
        Factory method that converts a geometry.Point object into a Pt object.

        Args:
            point (geometry.Point): point to be converted into Pt object.

        Returns:
            Pt: the returned Pt object.
        """
        return Pt(point.x, point.y)

    def as_geo_pt(self) -> geo_pt.GeoPt:
        """
        Converts the Pt into a GeoPt object.

        Returns:
            GeoPt: the converted point.
        """
        return geo_pt.GeoPt(self.y, self.x)

    def coords_as_tuple_xy(self) -> Tuple[float, float]:
        """
        Converts the point into an x-y tuple

        Returns:
            Tuple[float, float]: x-y coordinate tuple.
        """
        return (self.x, self.y)

    def coords_as_tuple_yx(self) -> Tuple[float, float]:
        """
        Converts the point into a y-x tuple

        Returns:
            Tuple[float, float]: y-x coordinate tuple.
        """
        return (self.y, self.x)

    def get_distance(self, point: Pt) -> float:
        """
        Computes the distance to another Pt object.
        Does not allow computing distance to GeoPt objects as they involve different computations.

        Args:
            point (Pt): target point to be computed.

        Raises:
            TypeError: a GeoPt was sent in as the point argument.

        Returns:
            float: distance, in units, to the target point.
        """
        if isinstance(point, Pt):
            raise TypeError("Cannot be GeoPt!")
        return ((self.y-point.y)**2+(self.x-point.x)**2)**0.5
    
    def get_closest_point(self, *points: Pt) -> Tuple[Optional[Pt], float]:
        """
        Based on a collection of points, find the nearest to self.
        
        Args:
            *points (Pt): we will find the closest of these points to self.

        Returns:
            Tuple[Optional[Pt], float]: point-distance tuple.
        """
        nearest_point = None
        nearest_dist = float("inf")
        for point in points:
            dist = self.get_distance(point)
            if nearest_dist > dist:
                nearest_dist = dist
                nearest_point = point
        return (nearest_point, nearest_dist)

    def move_to(self, new_x: float, new_y: float) -> Pt:
        return Pt(new_x, new_y)


from __future__ import annotations
from typing import Any, Callable, Dict, Generic, List, Optional, Set, Tuple, TypeVar

from .comparable import Comparable

T = TypeVar("T", bound=Comparable)
U = TypeVar("U")

class Node(Generic[T]):
    """
    Encapsulates a Node of the AVLTree.
    
    Fields:
        key (T): the key to be assigned to the node.
        left (Optional[Node[T]]): the left child of the node.
        right (Optional[Node[T]]): the right child of the node.
        height (int): the height of the tree rooted at the node.
        count (int): duplicate values are captured here, so we can have nodes with >1 counts.
        weight (int): the number of nodes on the tree rooted at the node.
    """
    key:    T
    left:   Optional[Node[T]]
    right:  Optional[Node[T]]
    height: int
    count:  int
    weight: int

    def __init__(self, key: T):
        self.key = key
        self.left = None
        self.right = None
        self.height = 1
        self.count = 1
        self.weight = 1

class AVLTree(Generic[U, T]):
    """
    Encapsulates an AVLTree capable of performing rotations, insertions and deletions.
    Each node is represented by a key and can contain many values in the form of a list.
    For example, if we want to contain the scores of students in a class,
        the key for comparison would be scores, while each node would contain a list of students.
        T would be int, and U would be Student in this case.

    Fields:
        root (Optional[Node[T]]): the root of the tree.
    """
    root: Optional[Node[T]]
    keys: Dict[T, List[U]]
    comparator: Callable[[U], T]
            
    def __init__(self, comparator: Callable[[U], T]):
        self.root = None
        self.keys = {}
        self.comparator = comparator
    
    def insert(self, value: U) -> None:
        """
        Inserts a value to the tree, triggering rotations.

        Args:
            value (U): value to be added.
        """
        key = self.comparator(value)
        if key not in self.keys:
            self.keys[key] = [value]
        else:
            self.keys[key].append(value)
        def insert_helper(node: Optional[Node[T]], key: T) -> Optional[Node[T]]:
            if not node:
                return Node[T](key)
            elif key < node.key:
                node.left = insert_helper(node.left, key)
            elif key > node.key:
                node.right = insert_helper(node.right, key)
            else:
                node.count += 1
            self.set_height(node)
            self.set_weight(node)
            
            balance = self.get_balance(node)
            
            left_key = self.get_key(node.left)
            right_key = self.get_key(node.right)

            if balance > 1 and key < left_key:
                return self.right_rotate(node)
            if balance < -1 and right_key and key > right_key:
                return self.left_rotate(node)
            if balance > 1 and left_key and key > left_key:
                node.left = self.left_rotate(node.left)
                return self.right_rotate(node)
            if balance < -1 and key < right_key:
                node.right = self.right_rotate(node.right)
                return self.left_rotate(node)

            return node
        self.root = insert_helper(self.root, key)

    def delete(self, value: Optional[U]) -> None:
        """
        Deletes a key from the tree, triggering rotations.

        Args:
            value (U): value to be deleted.
        """
        def drop_child(node: Optional[Node[T]], key: T) -> bool:
            if not node:
                return False
            elif key == node.key:
                return True
            elif key < node.key:
                if drop_child(node.left, key):
                    node.left = None
                return False
            elif key > node.key:
                if drop_child(node.right, key):
                    node.right = None
                return False
            return False
            
        def delete_helper(node: Optional[Node[T]], key: T) -> Optional[Node[T]]:
            if not node:
                return None
            elif key < node.key:
                node.left = delete_helper(node.left, key)
            elif key > node.key:
                node.right = delete_helper(node.right, key)
            elif node.count > 1:
                node.count -= 1
            else:
                if node.left is None:
                    temp = node.right
                    node = None
                    return temp
                elif node.right is None:
                    temp = node.left
                    node = None
                    return temp
                temp = self.get_min_value_node(node.right)
                node.key = temp.key
                node.count = temp.count
                node.right = delete_helper(node.right, temp.key)
            if not node:
                return node
            self.set_height(node)
            self.set_weight(node)
            balance = self.get_balance(node)

            if balance > 1 and self.get_balance(node.left) >= 0:
                return self.right_rotate(node)
            if balance < -1 and self.get_balance(node.right) <= 0:
                return self.left_rotate(node)
            if balance > 1 and self.get_balance(node.left) < 0:
                node.left = self.left_rotate(node.left)
                return self.right_rotate(node)
            if balance < -1 and self.get_balance(node.right) > 0:
                node.right = self.right_rotate(node.right)
                return self.left_rotate(node)
            return node
        key = self.comparator(value)
        if key in self.keys and value in self.keys[key]:
            self.keys[key].remove(value)
            self.root = delete_helper(self.root, key)
        else:
            return
 
    def left_rotate(self, node: Optional[Node[T]]) -> Optional[Node[T]]:
        """
        Rotates the node left.

        Args:
            node (Optional[Node[T]]): node to be rotated.

        Returns:
            Optional[Node[T]]: the resulting root node.
        """
        if not node or not node.right:
            return node
        right = node.right
        right_left = right.left
        right.left = node
        node.right = right_left
 
        self.set_height(node)
        self.set_weight(node)
        self.set_height(right)
        self.set_weight(right)
        return right
 
    def right_rotate(self, node: Optional[Node[T]]) -> Optional[Node[T]]:
        """
        Rotates the node right.

        Args:
            node (Optional[Node[T]]): node to be rotated.

        Returns:
            Optional[Node[T]]: the resulting root node.
        """
        if not node or not node.left:
            return node
        left = node.left
        left_right = left.right
        left.right = node
        node.left = left_right
 
        self.set_height(node)
        self.set_weight(node)
        self.set_height(left)
        self.set_weight(left)
        return left
    
    def get_key(self, node: Optional[Node[T]]) -> Optional[T]:
        if not node:
            return None
        return node.key
 
    def get_height(self, node: Optional[Node[T]]) -> int:
        if not node:
            return 0
        return node.height
    
    def set_height(self, node: Node[T]) -> None:
        node.height = 1 + max(self.get_height(node.left), self.get_height(node.right))
        
    def get_weight(self, node: Optional[Node[T]]) -> int:
        if not node:
            return 0
        return node.weight
    
    def set_weight(self, node: Node[T]) -> None:
        node.weight = node.count + self.get_weight(node.left) + self.get_weight(node.right)
 
    def get_balance(self, node: Optional[Node[T]]) -> int:
        if not node:
            return 0
        return self.get_height(node.left) - self.get_height(node.right)
    
    def get_min_value(self) -> Optional[U]:
        """
        Gets the minimum value of the tree.

        Returns:
            T: the minimum value.
        """
        def get_min_value_helper(node: Optional[Node[T]]) -> Optional[U]:
            if not node:
                return None
            elif not node.left:
                return self.keys[node.key][-1]
            return get_min_value_helper(node.left)
        return get_min_value_helper(self.root)
    
    def get_min_value_node(self, node: Node[T]) -> Node[T]:
        """
        Get the node containing the minimum value in the tree rooted at this node.

        Args:
            node (Node[T]): node to be queried.

        Returns:
            Node: node containing the minimum value.
        """
        if node is None or node.left is None:
            return node
        return self.get_min_value_node(node.left)

    def in_order(self, accessor: Callable[[U], Any]=None) -> List[Any]:
        """
        In order traversal of the tree done by recursion.

        Returns:
            List[T]: a list of keys in the tree in ascending order.
        """
        if accessor == None:
            accessor = self.comparator
        L = []
        def in_order_helper(node: Optional[Node[T]]) -> None:
            if not node:
                return
            in_order_helper(node.left)
            L.extend([accessor(value) for value in self.keys[node.key]])
            in_order_helper(node.right)
        in_order_helper(self.root)
        return L
    
    def pre_order(self, accessor: Callable[[U], Any]=None) -> List[Any]:
        """
        Pre order traversal of the tree done by recursion.

        Returns:
            List[T]: a list of keys in the tree in pre order.
        """
        if accessor == None:
            accessor = self.comparator
        L = []
        def pre_order_helper(node: Optional[Node[T]]) -> None:
            if not node:
                return
            L.extend([accessor(value) for value in self.keys[node.key]])
            pre_order_helper(node.left)
            pre_order_helper(node.right)
        pre_order_helper(self.root)
        return L
    
    def get_rank(self, key: T) -> int:
        """
        Gets the rank of a particular key in the tree.
        Does take into account duplicate values.

        Args:
            key (T): key to be queried.

        Returns:
            int: its rank, starting from 1.
        """
        def rank_helper(node: Optional[Node[T]], key: T) -> int:
            if not node:
                return 0
            elif key < node.key:
                return rank_helper(node.left, key)
            elif key > node.key:
                return self.get_weight(node) - self.get_weight(node.right) + rank_helper(node.right, key)
            return self.get_weight(node.left)
        return rank_helper(self.root, key) + 1
    